"""Tangent Hyperbolic."""
import numpy

from ..baseclass import Dist
from .. import evaluation, approximation


class Tanh(Dist):
    """
    Tangent Hyperbolic.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Tanh(chaospy.Uniform(0, 1))
        >>> print(distribution)
        Tanh(Uniform(lower=0, upper=1))
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.1974 0.3799 0.537  0.664 ]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [1.0201 1.0811 1.1855 1.3374]
        >>> print(numpy.around(distribution.sample(4), 4))
        [0.5741 0.1145 0.7399 0.448 ]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.4065
    """

    def __init__(self, dist):
        assert isinstance(dist, Dist)
        assert numpy.all(dist.range() <= 1)
        assert numpy.all(dist.range() >= -1)
        Dist.__init__(self, dist=dist)

    def _pdf(self, x, dist, cache):
        """Probability density function."""
        return evaluation.evaluate_density(
            dist, numpy.arctanh(x), cache=cache)/numpy.sqrt(1-x*x)

    def _cdf(self, x, dist, cache):
        return evaluation.evaluate_forward(dist, numpy.arctanh(x), cache=cache)

    def _ppf(self, q, dist, cache):
        return numpy.tanh(evaluation.evaluate_inverse(dist, q, cache=cache))

    def _bnd(self, x, dist, cache):
        return numpy.tanh(evaluation.evaluate_bound(
            dist, numpy.arctanh(x), cache=cache))

    def _mom(self, x, dist, cache):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"



"""Arc-Sinus."""
import numpy
from ..baseclass import Dist
from .. import evaluation, approximation


class Arcsin(Dist):
    """
    Arc-Sinus.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Arcsin(chaospy.Uniform(0, 1))
        >>> print(distribution)
        Arcsin(Uniform(lower=0, upper=1))
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.2014 0.4115 0.6435 0.9273]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.9798 0.9165 0.8    0.6   ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [0.7123 0.1153 1.2541 0.5032]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.5708
        >>> print(numpy.around(distribution.ttr([0, 1, 2]), 4))
        [[0.5708 0.7302 0.7625]
         [1.     0.1416 0.1492]]
    """

    def __init__(self, dist):
        assert isinstance(dist, Dist)
        Dist.__init__(self, dist=dist)

    def _pdf(self, x, dist, cache):
        """Probability density function."""
        return evaluation.evaluate_density(
            dist, numpy.sin(x), cache=cache)*numpy.cos(x)

    def _cdf(self, x, dist, cache):
        return evaluation.evaluate_forward(dist, numpy.sin(x), cache=cache)

    def _ppf(self, q, dist, cache):
        return numpy.arcsin(evaluation.evaluate_inverse(dist, q, cache=cache))

    def _bnd(self, x, dist, cache):
        return numpy.arcsin(evaluation.evaluate_bound(
            dist, numpy.sin(x), cache=cache))

    def _mom(self, x, dist, cache):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"

    def _fwd_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.arcsin(dist)
        return self

    def _inv_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.sin(dist)
        return self



"""
Implementation of Bastos-O'Hagen algorithm for modified Cholesky decomposition.

Algorithm 3 from "Pivoting Cholesky Decomposition applied to Emulation and
Validation of computer models" by lowtri.S. Bastos and and mat. O'Hagan, 2007
"""
import numpy


def bastos_ohagen(mat, eps=1e-16):
    """
    Bastos-O'Hagen algorithm for modified Cholesky decomposition.

    Args:
        mat (numpy.ndarray):
            Input matrix to decompose. Assumed to close to positive definite.
        eps (float):
            Tolerance value for the eigenvalues. Values smaller
            than `tol*numpy.diag(mat).max()` are considered to be zero.

    Returns:
        (:py:data:typing.Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]):
            perm:
                Permutation matrix
            lowtri:
                Upper triangular decomposition
            errors:
                Error matrix

    Examples:
        >>> mat = [[4, 2, 1], [2, 6, 3], [1, 3, -.004]]
        >>> perm, lowtri = bastos_ohagen(mat)
        >>> print(perm)
        [[0 1 0]
         [1 0 0]
         [0 0 1]]
        >>> print(numpy.around(lowtri, 4))
        [[ 2.4495  0.      0.    ]
         [ 0.8165  1.8257  0.    ]
         [ 1.2247 -0.      0.9129]]
        >>> comp = numpy.dot(perm, lowtri)
        >>> print(numpy.around(numpy.dot(comp, comp.T), 4))
        [[4.     2.     1.    ]
         [2.     6.     3.    ]
         [1.     3.     2.3333]]
    """
    mat_ref = numpy.asfarray(mat)
    mat = mat_ref.copy()
    diag_max = numpy.diag(mat).max()
    assert len(mat.shape) == 2
    size = len(mat)

    hitri = numpy.zeros((size, size))
    piv = numpy.arange(size)

    for idx in range(size):

        idx_max = numpy.argmax(numpy.diag(mat[idx:, idx:])) + idx

        if mat[idx_max, idx_max] <= numpy.abs(diag_max*eps):

            if not idx:
                raise ValueError("Purly negative definite")

            for j in range(idx, size):
                hitri[j, j] = hitri[j-1, j-1]/float(j)

            break

        tmp = mat[:, idx].copy()
        mat[:, idx] = mat[:, idx_max]
        mat[:, idx_max] = tmp
        tmp = hitri[:, idx].copy()
        hitri[:, idx] = hitri[:, idx_max]
        hitri[:, idx_max] = tmp
        tmp = mat[idx, :].copy()
        mat[idx, :] = mat[idx_max, :]
        mat[idx_max, :] = tmp
        piv[idx], piv[idx_max] = piv[idx_max], piv[idx]

        hitri[idx, idx] = numpy.sqrt(mat[idx, idx])
        rval = mat[idx, idx+1:]/hitri[idx, idx]
        hitri[idx, idx+1:] = rval
        mat[idx+1:, idx+1:] -= numpy.outer(rval, rval)

    perm = numpy.zeros((size, size), dtype=int)
    for idx in range(size):
        perm[idx, piv[idx]] = 1

    return perm, hitri.T



"""Log-Normal probability distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class log_normal(Dist):

    def __init__(self, a=1):
        Dist.__init__(self, a=a)

    def _pdf(self, x, a):
        out = numpy.e**(-numpy.log(x+(1-x)*(x<=0))**2/(2*a*a)) / \
            ((x+(1-x)*(x<=0))*a*numpy.sqrt(2*numpy.pi))*(x>0)
        return out

    def _cdf(self, x, a):
        return special.ndtr(numpy.log(x+(1-x)*(x<=0))/a)*(x>0)

    def _ppf(self, x, a):
        return numpy.e**(a*special.ndtri(x))

    def _mom(self, k, a):
        return numpy.e**(.5*a*a*k*k)

    def _ttr(self, n, a):
        """Stieltjes-Wigert."""
        return \
    (numpy.e**(n*a*a)*(numpy.e**(a*a)+1)-1)*numpy.e**(.5*(2*n-1)*a*a), \
                (numpy.e**(n*a*a)-1)*numpy.e**((3*n-2)*a*a)

    def _bnd(self, x, a):
        return 0, self._ppf(1-1e-10, a)


class LogNormal(Add):
    R"""
    Log-normal distribution

    Args:
        mu (float, Dist) : Mean in the normal distribution.  Overlaps with
                scale by mu=log(scale)
        sigma (float, Dist) : Standard deviation of the normal distribution.
        shift (float, Dist) : Location of the lower bound.
        scale (float, Dist) : Scale parameter. Overlaps with mu by scale=e**mu

    Examples:
        >>> distribution = chaospy.LogNormal(0, 1)
        >>> print(distribution)
        LogNormal(mu=0, scale=1, shift=0, sigma=1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.431  0.7762 1.2883 2.3201]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.6495 0.4977 0.2999 0.1207]
        >>> print(numpy.around(distribution.sample(4), 4))
        [1.4844 0.3011 5.1945 0.9563]
        >>> print(numpy.around(distribution.mom(1), 4))
        1.6487
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[1.50155000e+01 1.18650900e+02 8.97651100e+02]
         [4.67080000e+00 3.48830600e+02 2.09298326e+04]]
    """

    def __init__(self, mu=0, sigma=1, shift=0, scale=1):
        self._repr = {"mu": mu, "sigma": sigma, "shift": shift, "scale": scale}
        left = log_normal(sigma)*scale*numpy.e**mu
        Add.__init__(self, left=left, right=shift)


Lognormal = deprecation_warning(LogNormal, "Lognormal")


class Gilbrat(Add):
    """
    Gilbrat distribution.

    Standard log-normal distribution

    Args:
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Examples:
        >>> distribution = chaospy.Gilbrat(3, 2)
        >>> print(distribution)
        Gilbrat(scale=3, shift=2)
        >>> q = numpy.linspace(0, 1, 6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [3.293  4.3286 5.865  8.9604]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.2165 0.1659 0.1    0.0402]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 6.4533  2.9033 17.5835  4.869 ]
        >>> print(numpy.around(distribution.mom(1), 4))
        6.9462
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[4.70464000e+01 3.57952700e+02 2.69495320e+03]
         [4.20370000e+01 3.13947580e+03 1.88368494e+05]]
    """
    def __init__(self, scale=1, shift=0):
        self._repr = {"scale": scale, "shift": shift}
        Add.__init__(self, left=log_normal(1)*scale, right=shift)



"""Power normal or Box-Cox distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class power_normal(Dist):
    """Power normal or Box-Cox distribution."""

    def __init__(self, c):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        norm = (2*numpy.pi)**-.5*numpy.exp(-x**2/2.)
        return c*norm*special.ndtr(-x)**(c-1.)

    def _cdf(self, x, c):
        return 1.-special.ndtr(-x)**c

    def _ppf(self, q, c):
        return -special.ndtri(pow(1-q, 1./c))

    def _bnd(self, x, c):
        return self._ppf(1e-10, c), self._ppf(1-1e-10, c)


class PowerNormal(Add):
    """
    Power normal or Box-Cox distribution.

    Args:
        shape (float, Dist) : Shape parameter
        mu (float, Dist) : Mean of the normal distribution
        scale (float, Dist) : Standard deviation of the normal distribution

    Examples:
        >>> distribution = chaospy.PowerNormal(2, 2, 2)
        >>> print(distribution)
        PowerNormal(mu=2, scale=2, shape=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [-0.5008  0.4919  1.3233  2.2654]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.1633 0.2325 0.2383 0.1768]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 1.5523 -1.122   3.5244  0.8368]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.8716
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[0.6455 0.4421 0.2628]
         [2.7268 5.5707 8.4597]]
    """

    def __init__(self, shape=1, mu=0, scale=1):
        self._repr = {"shape": shape, "mu": mu, "scale": scale}
        Add.__init__(self, left=power_normal(shape)*scale, right=mu)


Powernorm = deprecation_warning(PowerNormal, "Powernorm")



"""
Truncation.

Example usage
-------------

    >>> distribution = chaospy.Normal(0, 1)
    >>> print(numpy.around(distribution.inv([0.9, 0.99, 0.999]), 4))
    [1.2816 2.3263 3.0902]
    >>> distribution = chaospy.Normal(0, 1) < 1
    >>> print(numpy.around(distribution.inv([0.9, 0.99, 0.999]), 4))
    [0.6974 0.9658 0.9965]
    >>> distribution = chaospy.Normal(0, 1) < 2
    >>> print(numpy.around(distribution.inv([0.9, 0.99, 0.999]), 4))
    [1.1726 1.8449 1.9822]

Illegal dependencies:

    >>> dist1 = chaospy.Uniform()
    >>> dist2 = chaospy.Trunc(dist1, 0.5)
    >>> dist = chaospy.J(dist1, dist2)
    >>> dist.sample() # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    chaospy.distributions.baseclass.StochasticallyDependentError: truncated variable ...
"""
import numpy

from .joint import J
from ..baseclass import Dist, StochasticallyDependentError
from .. import evaluation


class Trunc(Dist):
    """Truncation."""

    def __init__(self, left, right):
        """
        Constructor.

        Args:
            left (Dist, numpy.ndarray) : Left hand side.
            right (Dist, numpy.ndarray) : Right hand side.
        """
        if isinstance(left, Dist) and len(left) > 1:
            if (not isinstance(left, J) or
                    evaluation.get_dependencies(*list(left.inverse_map))):
                raise StochasticallyDependentError(
                    "Joint distribution with dependencies not supported.")
        if isinstance(right, Dist) and len(right) > 1:
            if (not isinstance(right, J) or
                    evaluation.get_dependencies(*list(right.inverse_map))):
                raise StochasticallyDependentError(
                    "Joint distribution with dependencies not supported.")

        assert isinstance(left, Dist) or isinstance(right, Dist)
        Dist.__init__(self, left=left, right=right)

    def _bnd(self, xloc, left, right, cache):
        """
        Distribution bounds.

        Example:
            >>> print(chaospy.Uniform().range([-2, 0, 2, 4]))
            [[0. 0. 0. 0.]
             [1. 1. 1. 1.]]
            >>> print(chaospy.Trunc(chaospy.Uniform(), 0.6).range([-2, 0, 2, 4]))
            [[0.  0.  0.  0. ]
             [0.6 0.6 0.6 0.6]]
            >>> print(chaospy.Trunc(0.4, chaospy.Uniform()).range([-2, 0, 2, 4]))
            [[0.4 0.4 0.4 0.4]
             [1.  1.  1.  1. ]]
        """
        if isinstance(left, Dist):
            if left in cache:
                left = cache[left]
            else:
                left = evaluation.evaluate_bound(left, xloc, cache=cache)
        else:
            left = (numpy.array(left).T * numpy.ones((2,)+xloc.shape).T).T

        if isinstance(right, Dist):
            if right in cache:
                right = cache[right]
            else:
                right = evaluation.evaluate_bound(right, xloc, cache=cache)
        else:
            right = (numpy.array(right).T * numpy.ones((2,)+xloc.shape).T).T

        return left[0], right[1]

    def _cdf(self, xloc, left, right, cache):
        """
        Cumulative distribution function.

        Example:
            >>> print(chaospy.Uniform().fwd([-0.5, 0.3, 0.7, 1.2]))
            [0.  0.3 0.7 1. ]
            >>> print(chaospy.Trunc(chaospy.Uniform(), 0.4).fwd([-0.5, 0.2, 0.8, 1.2]))
            [0.  0.5 1.  1. ]
            >>> print(chaospy.Trunc(0.6, chaospy.Uniform()).fwd([-0.5, 0.2, 0.8, 1.2]))
            [0.  0.  0.5 1. ]
        """
        if isinstance(left, Dist) and left in cache:
            left = cache[left]
        if isinstance(right, Dist) and right in cache:
            right = cache[right]

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise StochasticallyDependentError(
                    "under-defined distribution {} or {}".format(left, right))
        else:
            left = (numpy.array(left).T*numpy.ones(xloc.shape).T).T
            uloc1 = evaluation.evaluate_forward(right, left, cache=cache.copy())
            uloc2 = evaluation.evaluate_forward(right, xloc, cache=cache)
            return (uloc2-uloc1)/(1-uloc1)

        right = (numpy.array(right).T*numpy.ones(xloc.shape).T).T
        uloc1 = evaluation.evaluate_forward(left, right, cache=cache.copy())
        uloc2 = evaluation.evaluate_forward(left, xloc, cache=cache)
        return uloc2/uloc1

    def _pdf(self, xloc, left, right, cache):
        """
        Probability density function.

        Example:
            >>> dist = chaospy.Trunc(chaospy.Uniform(), 0.6)
            >>> print(dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25]))
            [0.         1.66666667 1.66666667 0.         0.        ]
            >>> dist = chaospy.Trunc(chaospy.Uniform(), 0.4)
            >>> print(dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25]))
            [0.  2.5 0.  0.  0. ]
            >>> dist = chaospy.Trunc(0.4, chaospy.Uniform())
            >>> print(dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25]))
            [0.         0.         1.66666667 1.66666667 0.        ]
            >>> dist = chaospy.Trunc(0.6, chaospy.Uniform())
            >>> print(dist.pdf([-0.25, 0.25, 0.5, 0.75, 1.25]))
            [0.  0.  0.  2.5 0. ]
        """
        if isinstance(left, Dist) and left in cache:
            left = cache[left]
        if isinstance(right, Dist) and right in cache:
            right = cache[right]

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise StochasticallyDependentError(
                    "under-defined distribution {} or {}".format(left, right))
        else:
            left = (numpy.array(left).T*numpy.ones(xloc.shape).T).T
            uloc1 = evaluation.evaluate_forward(right, left, cache=cache.copy())
            uloc2 = evaluation.evaluate_density(right, xloc, cache=cache)
            return uloc2/(1-uloc1)

        right = (numpy.array(right).T*numpy.ones(xloc.shape).T).T
        uloc1 = evaluation.evaluate_forward(left, right, cache=cache.copy())
        uloc2 = evaluation.evaluate_density(left, xloc, cache=cache)
        return uloc2/uloc1


    def _ppf(self, q, left, right, cache):
        """
        Point percentile function.

        Example:
            >>> print(chaospy.Uniform().inv([0.1, 0.2, 0.9]))
            [0.1 0.2 0.9]
            >>> print(chaospy.Trunc(chaospy.Uniform(), 0.4).inv([0.1, 0.2, 0.9]))
            [0.04 0.08 0.36]
            >>> print(chaospy.Trunc(0.6, chaospy.Uniform()).inv([0.1, 0.2, 0.9]))
            [0.64 0.68 0.96]
        """
        if isinstance(left, Dist) and left in cache:
            left = cache[left]
        if isinstance(right, Dist) and right in cache:
            right = cache[right]

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise StochasticallyDependentError(
                    "under-defined distribution {} or {}".format(left, right))
        elif not isinstance(right, Dist):
            raise StochasticallyDependentError(
                "truncated variable indirectly depends on underlying variable")
        else:
            left = (numpy.array(left).T*numpy.ones(q.shape).T).T
            uloc = evaluation.evaluate_forward(right, left)
            return evaluation.evaluate_inverse(right, q*(1-uloc)+uloc, cache=cache)

        right = (numpy.array(right).T*numpy.ones(q.shape).T).T
        uloc = evaluation.evaluate_forward(left, right, cache=cache.copy())
        return evaluation.evaluate_inverse(left, q*uloc, cache=cache)

    def __str__(self):
        return (self.__class__.__name__ + "(" + str(self.prm["left"]) +
                ", " + str(self.prm["right"]) + ")")

def Trunk(left, right):
    return Trunc(left, right)

def trunk(left, right):
    return Trunc(left, right)



"""Multivariate Log-Normal Distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from .deprecate import deprecation_warning


class MvLogNormal(Dist):
    """
    Multivariate Log-Normal Distribution.

    Args:
        loc (float, Dist): Mean vector
        scale (float, Dist): Covariance matrix or variance vector if scale
            is a 1-d vector.

    Examples:
        >>> distribution = chaospy.MvLogNormal([1, 2], [[1, 0.6], [0.6, 1]])
        >>> print(distribution)
        MvLogNormal(loc=[1.0, 2.0], scale=[[1.0, 0.6], [0.6, 1.0]])
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> print(numpy.around(distribution.inv(mesh), 4))
        [[[ 1.3847  2.7183  5.3361]
          [ 1.3847  2.7183  5.3361]
          [ 1.3847  2.7183  5.3361]]
        <BLANKLINE>
         [[ 2.874   4.3077  6.4566]
          [ 4.9298  7.3891 11.075 ]
          [ 8.4562 12.6745 18.9971]]]
        >>> print(numpy.around(distribution.fwd(distribution.inv(mesh)), 4))
        [[[0.25 0.5  0.75]
          [0.25 0.5  0.75]
          [0.25 0.5  0.75]]
        <BLANKLINE>
         [[0.25 0.25 0.25]
          [0.5  0.5  0.5 ]
          [0.75 0.75 0.75]]]
        >>> print(numpy.around(distribution.pdf(distribution.inv(mesh)), 4))
        [[0.0317 0.0135 0.0037]
         [0.0232 0.0099 0.0027]
         [0.0108 0.0046 0.0012]]
        >>> print(numpy.around(distribution.sample(4), 4))
        [[ 4.0351  0.8185 14.1201  2.5996]
         [23.279   1.8986  4.9261  5.8399]]
        >>> print(numpy.around(distribution.mom((1, 2)), 4))
        6002.9122
    """

    def __init__(self, loc=[0, 0], scale=[[1, 0.5], [0.5, 1]]):
        loc = numpy.asfarray(loc)
        scale = numpy.asfarray(scale)
        assert len(loc) == len(scale)
        self._repr = {"loc": str(loc.tolist()), "scale": str(scale.tolist())}

        C = numpy.linalg.cholesky(scale)
        Ci = numpy.linalg.inv(C)
        Dist.__init__(self, loc=loc, C=C, Ci=Ci, scale=scale)

    def _cdf(self, x, loc, C, Ci, scale):
        y = numpy.log(numpy.abs(x) + 1.*(x<=0))
        out = special.ndtr(numpy.dot(Ci, (y.T-loc.T).T))
        return numpy.where(x <= 0, 0., out)

    def _ppf(self, q, loc, C, Ci, scale):
        return numpy.e**(numpy.dot(C, special.ndtri(q)).T+loc.T).T

    def _mom(self, k, loc, C, Ci, scale):
        output =  numpy.dot(k, loc)
        output += .5*numpy.dot(numpy.dot(k, scale), k)
        output =  numpy.e**(output)
        return output

    def _bnd(self, x, loc, C, Ci, scale):
        up = (7.1*numpy.sqrt(numpy.diag(scale))*x.T**0 + loc.T).T
        return 0*up, numpy.e**up

    def __len__(self):
        return len(self.prm["C"])


MvLognormal = deprecation_warning(MvLogNormal, "MvLognormal")



"""
Collection of tools for manipulation polynomial.
"""

import numpy
from scipy.special import comb, factorial

import chaospy.poly.dimension
import chaospy.bertran
import chaospy.quadrature

from chaospy.poly.base import Poly


def basis(start, stop=None, dim=1, sort="G", cross_truncation=1.):
    """
    Create an N-dimensional unit polynomial basis.

    Args:
        start (int, numpy.ndarray):
            the minimum polynomial to include. If int is provided, set as
            lowest total order.  If array of int, set as lower order along each
            axis.
        stop (int, numpy.ndarray):
            the maximum shape included. If omitted:
            ``stop <- start; start <- 0`` If int is provided, set as largest
            total order. If array of int, set as largest order along each axis.
        dim (int):
            dim of the basis. Ignored if array is provided in either start or
            stop.
        sort (str):
            The polynomial ordering where the letters ``G``, ``I`` and ``R``
            can be used to set grade, inverse and reverse to the ordering.  For
            ``basis(start=0, stop=2, dim=2, order=order)`` we get:
            ======  ==================
            order   output
            ======  ==================
            ""      [1 y y^2 x xy x^2]
            "G"     [1 y x y^2 xy x^2]
            "I"     [x^2 xy x y^2 y 1]
            "R"     [1 x x^2 y xy y^2]
            "GIR"   [y^2 xy x^2 y x 1]
            ======  ==================
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion.

    Returns:
        (Poly) : Polynomial array.

    Examples:
        >>> print(chaospy.basis(4, 4, 2, sort="GR"))
        [q0^4, q0^3q1, q0^2q1^2, q0q1^3, q1^4]
        >>> print(chaospy.basis([1, 1], [2, 2], sort="GR"))
        [q0q1, q0^2q1, q0q1^2, q0^2q1^2]
    """
    if stop is None:
        start, stop = numpy.array(0), start

    start = numpy.array(start, dtype=int)
    stop = numpy.array(stop, dtype=int)
    dim = max(start.size, stop.size, dim)
    indices = numpy.array(chaospy.bertran.bindex(
        numpy.min(start), 2*numpy.max(stop), dim, sort, cross_truncation))

    if start.size == 1:
        bellow = numpy.sum(indices, -1) >= start

    else:
        start = numpy.ones(dim, dtype=int)*start
        bellow = numpy.all(indices-start >= 0, -1)


    if stop.size == 1:
        above = numpy.sum(indices, -1) <= stop.item()
    else:
        stop = numpy.ones(dim, dtype=int)*stop
        above = numpy.all(stop-indices >= 0, -1)

    pool = list(indices[above*bellow])

    arg = numpy.zeros(len(pool), dtype=int)
    arg[0] = 1
    poly = {}
    for idx in pool:
        idx = tuple(idx)
        poly[idx] = arg
        arg = numpy.roll(arg, 1)
    x = numpy.zeros(len(pool), dtype=int)
    x[0] = 1
    A = {}
    for I in pool:
        I = tuple(I)
        A[I] = x
        x = numpy.roll(x,1)

    return Poly(A, dim)


def lagrange(X):

    X = numpy.array(X)
    if len(X.shape) < 2:
        X = X.reshape(1, *X.shape)
    if len(X.shape) < 2:
        X = X.reshape(1, *X.shape)

    dim, K = X.shape

    return Poly(poly, dim)


def cutoff(poly, *args):
    """
    Remove polynomial components with order outside a given interval.

    Args:
        poly (Poly):
            Input data.
        low (int):
            The lowest order that is allowed to be included. Defaults to 0.
        high (int):
            The upper threshold for the cutoff range.

    Returns:
        (Poly):
            The same as `P`, except that all terms that have a order not within
            the bound `low <= order < high` are removed.

    Examples:
        >>> poly = chaospy.prange(4, 1) + chaospy.prange(4, 2)[::-1]
        >>> print(poly) # doctest: +SKIP
        [q1^3+1, q0+q1^2, q0^2+q1, q0^3+1]
        >>> print(chaospy.cutoff(poly, 3)) # doctest: +SKIP
        [1, q0+q1^2, q0^2+q1, 1]
        >>> print(chaospy.cutoff(poly, 1, 3)) # doctest: +SKIP
        [0, q0+q1^2, q0^2+q1, 0]
    """
    if len(args) == 1:
        low, high = 0, args[0]
    else:
        low, high = args[:2]

    core_old = poly.A
    core_new = {}
    for key in poly.keys:
        if low <= numpy.sum(key) < high:
            core_new[key] = core_old[key]
    return Poly(core_new, poly.dim, poly.shape, poly.dtype)


def dot(P, Q):

    P = Poly(P)
    Q = Poly(Q)
    if numpy.prod(P.shape)<=1 or numpy.prod(Q.shape)<=1:
        return P*Q
    return sum(P*Q, -1)


def differential(P, Q):
    """
    Polynomial differential operator.

    Args:
        P (Poly):
            Polynomial to be differentiated.
        Q (Poly):
            Polynomial to differentiate by. Must be decomposed. If polynomial
            array, the output is the Jacobian matrix.
    """
    P, Q = Poly(P), Poly(Q)

    if not chaospy.poly.is_decomposed(Q):
        differential(chaospy.poly.decompose(Q)).sum(0)

    if Q.shape:
        return Poly([differential(P, q) for q in Q])

    if Q.dim>P.dim:
        P = chaospy.poly.setdim(P, Q.dim)
    else:
        Q = chaospy.poly.setdim(Q, P.dim)

    qkey = Q.keys[0]

    A = {}
    for key in P.keys:

        newkey = numpy.array(key) - numpy.array(qkey)

        if numpy.any(newkey<0):
            continue

        A[tuple(newkey)] = P.A[key]*numpy.prod([factorial(key[i], \
            exact=True)/factorial(newkey[i], exact=True) \
            for i in range(P.dim)])

    return Poly(B, P.dim, P.shape, P.dtype)


def prange(N=1, dim=1):
    """
    Constructor to create a range of polynomials where the exponent vary.

    Args:
        N (int):
            Number of polynomials in the array.
        dim (int):
            The dimension the polynomial should span.

    Returns:
        (Poly):
            A polynomial array of length N containing simple polynomials with
            increasing exponent.

    Examples:
        >>> print(prange(4))
        [1, q0, q0^2, q0^3]
        >>> print(prange(4, dim=3))
        [1, q2, q2^2, q2^3]
    """
    A = {}
    r = numpy.arange(N, dtype=int)
    key = numpy.zeros(dim, dtype=int)
    for i in range(N):
        key[-1] = i
        A[tuple(key)] = 1*(r==i)

    return Poly(A, dim, (N,), int)


def rolldim(P, n=1):
    """
    Roll the axes.

    Args:
        P (Poly) : Input polynomial.
        n (int) : The axis that after rolling becomes the 0th axis.

    Returns:
        (Poly) : Polynomial with new axis configuration.

    Examples:
        >>> x,y,z = variable(3)
        >>> P = x*x*x + y*y + z
        >>> print(P)
        q0^3+q1^2+q2
        >>> print(rolldim(P))
        q0^2+q2^3+q1
    """
    dim = P.dim
    shape = P.shape
    dtype = P.dtype
    A = dict(((key[n:]+key[:n],P.A[key]) for key in P.keys))
    return Poly(A, dim, shape, dtype)


def swapdim(P, dim1=1, dim2=0):
    """
    Swap the dim between two variables.

    Args:
        P (Poly):
            Input polynomial.
        dim1 (int):
            First dim
        dim2 (int):
            Second dim.

    Returns:
        (Poly):
            Polynomial with swapped dimensions.

    Examples:
        >>> x,y = variable(2)
        >>> P = x**4-y
        >>> print(P)
        q0^4-q1
        >>> print(swapdim(P))
        q1^4-q0
    """
    if not isinstance(P, Poly):
        return numpy.swapaxes(P, dim1, dim2)

    dim = P.dim
    shape = P.shape
    dtype = P.dtype

    if dim1==dim2:
        return P

    m = max(dim1, dim2)
    if P.dim <= m:
        P = chaospy.poly.dimension.setdim(P, m+1)
        dim = m+1

    A = {}

    for key in P.keys:

        val = P.A[key]
        key = list(key)
        key[dim1], key[dim2] = key[dim2], key[dim1]
        A[tuple(key)] = val

    return Poly(A, dim, shape, dtype)


def tril(P, k=0):
    """Lower triangle of coefficients."""
    A = P.A.copy()
    for key in P.keys:
        A[key] = numpy.tril(P.A[key])
    return Poly(A, dim=P.dim, shape=P.shape)


def tricu(P, k=0):
    """Cross-diagonal upper triangle."""
    tri = numpy.sum(numpy.mgrid[[slice(0,_,1) for _ in P.shape]], 0)
    tri = tri<len(tri) + k

    if isinstance(P, Poly):
        A = P.A.copy()
        B = {}
        for key in P.keys:
            B[key] = A[key]*tri
        return Poly(B, shape=P.shape, dim=P.dim, dtype=P.dtype)

    out = P*tri
    return out


def variable(dims=1):
    """
    Simple constructor to create single variables to create polynomials.

    Args:
        dims (int):
            Number of dimensions in the array.

    Returns:
        (Poly):
            Polynomial array with unit components in each dimension.

    Examples:
        >>> print(variable())
        q0
        >>> print(variable(3))
        [q0, q1, q2]
    """
    if dims == 1:
        return Poly({(1,): 1}, dim=1, shape=())
    return Poly({
        tuple(indices): indices for indices in numpy.eye(dims, dtype=int)
    }, dim=dims, shape=(dims,))

def order(P):

    out = numpy.zeros(P.shape, dtype=int)
    for key in P.keys:
        o = sum(key)
        out = numpy.max([out, o*(P.A[key])], 0)
    return out


def all(A, ax=None):
    """Test if all values in A evaluate to True """
    if isinstance(A, Poly):
        out = numpy.zeros(A.shape, dtype=bool)
        B = A.A
        for key in A.keys:
            out += all(B[key], ax)
        return out

    return numpy.all(A, ax)


def any(A, ax=None):
    """Test if any values in A evaluate to True """
    if isinstance(A, Poly):
        out = numpy.zeros(A.shape, dtype=bool)
        B = A.A
        for key in A.keys:
            out *= any(B[key])
        return out

    return numpy.any(A, ax)


def around(A, decimals=0):
    """
    Evenly round to the given number of decimals.

    Args:
        A (Poly, numpy.ndarray):
            Input data.
        decimals (int):
            Number of decimal places to round to (default: 0).  If decimals is
            negative, it specifies the number of positions to the left of the
            decimal point.

    Returns:
        (Poly, numpy.ndarray):
            Same type as A.

    Examples:
        >>> P = chaospy.prange(3)*2**-numpy.arange(0, 6, 2, float)
        >>> print(P)
        [1.0, 0.25q0, 0.0625q0^2]
        >>> print(chaospy.around(P))
        [1.0, 0.0, 0.0]
        >>> print(chaospy.around(P, 2))
        [1.0, 0.25q0, 0.06q0^2]
    """
    if isinstance(A, Poly):
        B = A.A.copy()
        for key in A.keys:
            B[key] = around(B[key], decimals)
        return Poly(B, A.dim, A.shape, A.dtype)

    return numpy.around(A, decimals)


def diag(A, k=0):
    """Extract or construct a diagonal polynomial array."""
    if isinstance(A, Poly):
        core, core_new = A.A, {}
        for key in A.keys:
            core_new[key] = numpy.diag(core[key], k)

        return Poly(core_new, A.dim, None, A.dtype)

    return numpy.diag(A, k)


def repeat(A, repeats, axis=None):
    if isinstance(A, Poly):
        core = A.A.copy()
        for key in A.keys:
            core[key] = repeat(core[key], repeats, axis)
        return Poly(core, A.dim, None, A.dtype)

    return numpy.repeat(A, repeats, axis)


def trace(A, offset=0, ax1=0, ax2=1):
    if isinstance(A, Poly):
        core = A.A.copy()
        for key in A.keys:
            core[key] = trace(core[key], ax1, ax2)
        return Poly(core, A.dim, None, A.dtype)

    return numpy.trace(A, offset, ax1, ax2)


def prune(A, threshold):
    """
    Remove coefficients that is not larger than a given threshold.

    Args:
        A (Poly):
            Input data.
        threshold (float):
            Threshold for which values to cut.

    Returns:
        (Poly):
            Same type as A.

    Examples:
        >>> P = chaospy.sum(chaospy.prange(3)*2**-numpy.arange(0, 6, 2, float))
        >>> print(P)
        0.0625q0^2+0.25q0+1.0
        >>> print(chaospy.prune(P, 0.1))
        0.25q0+1.0
        >>> print(chaospy.prune(P, 0.5))
        1.0
        >>> print(chaospy.prune(P, 1.5))
        0.0
    """
    if isinstance(A, Poly):
        B = A.A.copy()
        for key in A.keys:
            values = B[key].copy()
            values[numpy.abs(values) < threshold] = 0.
            B[key] = values
        return Poly(B, A.dim, A.shape, A.dtype)

    A = A.copy()
    A[numpy.abs(A) < threshold] = 0.
    return A



"""Multivariate Normal Distribution."""
import numpy
from scipy import special

from chaospy.bertran import bindex
from .normal import normal

from ..baseclass import Dist


class MvNormal(Dist):
    """
    Multivariate Normal Distribution.

    Args:
        loc (float, Dist): Mean vector
        scale (float, Dist): Covariance matrix or variance vector if scale is a 1-d vector.

    Examples:
        >>> distribution = chaospy.MvNormal([1, 2], [[1, 0.6], [0.6, 1]])
        >>> print(distribution)
        MvNormal(loc=[1.0, 2.0], scale=[[1.0, 0.6], [0.6, 1.0]])
        >>> print(chaospy.Cov(distribution))
        [[1.  0.6]
         [0.6 1. ]]
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> print(numpy.around(distribution.inv(mesh), 4))
        [[[0.3255 1.     1.6745]
          [0.3255 1.     1.6745]
          [0.3255 1.     1.6745]]
        <BLANKLINE>
         [[1.0557 1.4604 1.8651]
          [1.5953 2.     2.4047]
          [2.1349 2.5396 2.9443]]]
        >>> print(numpy.around(distribution.fwd(distribution.inv(mesh)), 4))
        [[[0.25 0.5  0.75]
          [0.25 0.5  0.75]
          [0.25 0.5  0.75]]
        <BLANKLINE>
         [[0.25 0.25 0.25]
          [0.5  0.5  0.5 ]
          [0.75 0.75 0.75]]]
        >>> print(numpy.around(distribution.pdf(distribution.inv(mesh)), 4))
        [[0.0991 0.146  0.1452]
         [0.1634 0.1989 0.1634]
         [0.1452 0.146  0.0991]]
        >>> print(numpy.around(distribution.sample(4), 4))
        [[ 1.395  -0.2003  2.6476  0.9553]
         [ 3.1476  0.6411  1.5946  1.7647]]
        >>> print(numpy.around(distribution.mom((1, 2)), 4))
        7.4
    """

    def __init__(self, loc=[0, 0], scale=[[1, .5], [.5, 1]]):
        loc = numpy.asfarray(loc)
        scale = numpy.asfarray(scale)
        assert len(loc) == len(scale)
        self._repr = {"loc": loc.tolist(), "scale": scale.tolist()}

        C = numpy.linalg.cholesky(scale)
        Ci = numpy.linalg.inv(C)
        Dist.__init__(self, C=C, Ci=Ci, loc=loc)

    def _cdf(self, x, C, Ci, loc):
        return special.ndtr(numpy.dot(Ci, (x.T-loc.T).T))

    def _ppf(self, q, C, Ci, loc):
        return (numpy.dot(C, special.ndtri(q)).T+loc.T).T

    def _pdf(self, x, C, Ci, loc):
        det = numpy.linalg.det(numpy.dot(C,C.T))
        x_ = numpy.dot(Ci.T, (x.T-loc.T).T)
        out = numpy.ones(x.shape)
        out[0] =  numpy.e**(-.5*numpy.sum(x_*x_, 0))/numpy.sqrt((2*numpy.pi)**len(Ci)*det)
        return out

    def _bnd(self, x, C, Ci, loc):
        scale = numpy.sqrt(numpy.diag(numpy.dot(C,C.T)))
        lo,up = numpy.zeros((2,)+x.shape)
        lo.T[:] = (-7.5*scale+loc)
        up.T[:] = (7.5*scale+loc)
        return lo, up

    def _mom(self, k, C, Ci, loc):
        scale = numpy.dot(C, C.T)
        out = 0.
        for idx, kdx in enumerate(bindex(k, dim=len(C), sort="G")):
            coef = numpy.prod(special.comb(k.T, kdx).T, 0)
            diff = k.T - kdx
            pos = diff >= 0
            diff = diff*pos
            pos = numpy.all(pos)
            loc_ = numpy.prod(loc**diff)

            out += pos*coef*loc_*isserlis_moment(tuple(kdx), scale)

        return float(out)

    def __len__(self):
        return len(self.prm["C"])


def isserlis_moment(k, scale):
    """
    Raw statistical moments centralized Normal distribution using Isserlis'
    theorem.

    Args:
        k (Tuple[int, ...]):
            Moment orders.
        scale (ndarray):
            Covariance matrix defining dependencies between variables.

    Returns:
        Raw statistical moment of order ``k`` given covariance ``scale``.

    Examples:
        >>> scale = 0.5*numpy.eye(3)+0.5
        >>> isserlis_moment((2, 2, 2), scale)
        3.5
        >>> isserlis_moment((0, 0, 0), scale)
        1.0
        >>> isserlis_moment((1, 0, 0), scale)
        0.0
        >>> isserlis_moment((0, 1, 1), scale)
        0.5
        >>> isserlis_moment((0, 0, 2), scale)
        1.0
    """
    if not isinstance(k, numpy.ndarray):
        k = numpy.asarray(k)
    assert len(k.shape) == 1

    # Recursive exit condition
    if (numpy.sum(k) % 2 == 1) or numpy.any(k < 0):
        return 0.

    # Choose a pivot index as first non-zero entry
    idx = numpy.nonzero(k)[0]
    if not idx.size:
        # Skip if no pivot found
        return 1.
    idx = idx[0]

    eye = numpy.eye(len(k), dtype=int)
    out = (k[idx]-1)*scale[idx, idx]*isserlis_moment(k-2*eye[idx], scale)
    for idj in range(idx+1, len(k)):
        out += k[idj]*scale[idx, idj]*isserlis_moment(
            k-eye[idx]-eye[idj], scale)

    return float(out)



"""Power operator."""
import numpy

from ..baseclass import Dist, StochasticallyDependentError
from .. import evaluation


class Pow(Dist):
    """Power operator."""

    def __init__(self, left, right):
        """
        Constructor.

        Args:
            left (Dist, numpy.ndarray) : Left hand side.
            right (Dist, numpy.ndarray) : Right hand side.
        """
        Dist.__init__(self, left=left, right=right)

    def _bnd(self, xloc, left, right, cache):
        """
        Distribution bounds.

        Example:
            >>> print(chaospy.Uniform().range([-2, 0, 2, 4]))
            [[0. 0. 0. 0.]
             [1. 1. 1. 1.]]
            >>> print(chaospy.Pow(chaospy.Uniform(), 2).range([-2, 0, 2, 4]))
            [[0. 0. 0. 0.]
             [1. 1. 1. 1.]]
            >>> print(chaospy.Pow(chaospy.Uniform(1, 2), -1).range([-2, 0, 2, 4]))
            [[0.5 0.5 0.5 0.5]
             [1.  1.  1.  1. ]]
            >>> print(chaospy.Pow(2, chaospy.Uniform()).range([-2, 0, 2, 4]))
            [[1. 1. 1. 1.]
             [2. 2. 2. 2.]]
            >>> print(chaospy.Pow(2, chaospy.Uniform(-1, 0)).range([-2, 0, 2, 4]))
            [[0.5 0.5 0.5 0.5]
             [1.  1.  1.  1. ]]
            >>> print(chaospy.Pow(2, 3).range([-2, 0, 2, 4]))
            [[8. 8. 8. 8.]
             [8. 8. 8. 8.]]
        """
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise StochasticallyDependentError(
                    "under-defined distribution {} or {}".format(left, right))
        elif not isinstance(right, Dist):
            return left**right, left**right
        else:
            output = numpy.ones(xloc.shape)
            left = left * output
            assert numpy.all(left >= 0), "root of negative number"

            indices = xloc > 0
            output[indices] = numpy.log(xloc[indices])
            output[~indices] = -numpy.inf

            indices = left != 1
            output[indices] /= numpy.log(left[indices])

            output = evaluation.evaluate_bound(right, output, cache=cache)
            output = left**output
            output[:] = (
                numpy.where(output[0] < output[1], output[0], output[1]),
                numpy.where(output[0] < output[1], output[1], output[0]),
            )
            return output

        output = numpy.zeros(xloc.shape)
        right = right + output

        indices = right > 0
        output[indices] = numpy.abs(xloc[indices])**(1/right[indices])
        output[indices] *= numpy.sign(xloc[indices])
        output[right == 0] = 1
        output[(xloc == 0) & (right < 0)] = numpy.inf

        output = evaluation.evaluate_bound(left, output, cache=cache)

        pair = right % 2 == 0
        bnd_ = numpy.empty(output.shape)
        bnd_[0] = numpy.where(pair*(output[0]*output[1] < 0), 0, output[0])
        bnd_[0] = numpy.where(pair*(output[0]*output[1] > 0), \
                numpy.min(numpy.abs(output), 0), bnd_[0])**right
        bnd_[1] = numpy.where(pair, numpy.max(numpy.abs(output), 0),
                output[1])**right

        bnd_[0], bnd_[1] = (
            numpy.where(bnd_[0] < bnd_[1], bnd_[0], bnd_[1]),
            numpy.where(bnd_[0] < bnd_[1], bnd_[1], bnd_[0]),
        )
        return bnd_


    def _cdf(self, xloc, left, right, cache):
        """
        Cumulative distribution function.

        Example:
            >>> print(chaospy.Uniform().fwd([-0.5, 0.5, 1.5, 2.5]))
            [0.  0.5 1.  1. ]
            >>> print(chaospy.Pow(chaospy.Uniform(), 2).fwd([-0.5, 0.5, 1.5, 2.5]))
            [0.         0.70710678 1.         1.        ]
            >>> print(chaospy.Pow(chaospy.Uniform(1, 2), -1).fwd([0.4, 0.6, 0.8, 1.2]))
            [0.         0.33333333 0.75       1.        ]
            >>> print(chaospy.Pow(2, chaospy.Uniform()).fwd([-0.5, 0.5, 1.5, 2.5]))
            [0.        0.        0.5849625 1.       ]
            >>> print(chaospy.Pow(2, chaospy.Uniform(-1, 0)).fwd([0.4, 0.6, 0.8, 1.2]))
            [0.         0.26303441 0.67807191 1.        ]
            >>> print(chaospy.Pow(2, 3).fwd([7, 8, 9]))
            [0. 1. 1.]
        """
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise StochasticallyDependentError(
                    "under-defined distribution {} or {}".format(left, right))

        elif not isinstance(right, Dist):
            return numpy.inf

        else:
            assert numpy.all(left > 0), "imaginary result"

            y = (numpy.log(numpy.abs(xloc) + 1.*(xloc <= 0)) /
                 numpy.log(numpy.abs(left)+1.*(left == 1)))

            out = evaluation.evaluate_forward(right, y)
            out = numpy.where(xloc <= 0, 0., out)
            return out

        y = numpy.sign(xloc)*numpy.abs(xloc)**(1./right)
        pairs = numpy.sign(xloc**right) != -1

        out1, out2 = (
            evaluation.evaluate_forward(left, y, cache=cache),
            evaluation.evaluate_forward(left, -y, cache=cache),
        )
        out = numpy.where(right < 0, 1-out1, out1-pairs*out2)
        return out

    def _ppf(self, q, left, right, cache):
        """
        Point percentile function.

        Example:
            >>> print(chaospy.Uniform().inv([0.1, 0.2, 0.9]))
            [0.1 0.2 0.9]
            >>> print(chaospy.Pow(chaospy.Uniform(), 2).inv([0.1, 0.2, 0.9]))
            [0.01 0.04 0.81]
            >>> print(chaospy.Pow(chaospy.Uniform(1, 2), -1).inv([0.1, 0.2, 0.9]))
            [0.52631579 0.55555556 0.90909091]
            >>> print(chaospy.Pow(2, chaospy.Uniform()).inv([0.1, 0.2, 0.9]))
            [1.07177346 1.14869835 1.86606598]
            >>> print(chaospy.Pow(2, chaospy.Uniform(-1, 0)).inv([0.1, 0.2, 0.9]))
            [0.53588673 0.57434918 0.93303299]
            >>> print(chaospy.Pow(2, 3).inv([0.1, 0.2, 0.9]))
            [8. 8. 8.]
        """
        left = evaluation.get_inverse_cache(left, cache)
        right = evaluation.get_inverse_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise StochasticallyDependentError(
                    "under-defined distribution {} or {}".format(left, right))
        elif not isinstance(right, Dist):
            return left**right

        else:
            out = evaluation.evaluate_inverse(right, q, cache=cache)
            out = numpy.where(left < 0, 1-out, out)
            out = left**out
            return out

        right = right + numpy.zeros(q.shape)
        q = numpy.where(right < 0, 1-q, q)
        out = evaluation.evaluate_inverse(left, q, cache=cache)**right
        return out


    def _pdf(self, xloc, left, right, cache):
        """
        Probability density function.

        Example:
            >>> print(chaospy.Uniform().pdf([-0.5, 0.5, 1.5, 2.5]))
            [0. 1. 0. 0.]
            >>> print(chaospy.Pow(chaospy.Uniform(), 2).pdf([-0.5, 0.5, 1.5, 2.5]))
            [0.         0.70710678 0.         0.        ]
            >>> print(chaospy.Pow(chaospy.Uniform(1, 2), -1).pdf([0.4, 0.6, 0.8, 1.2]))
            [0.         2.77777778 1.5625     0.        ]
            >>> print(chaospy.Pow(2, chaospy.Uniform()).pdf([-0.5, 0.5, 1.5, 2.5]))
            [0.         0.         0.96179669 0.        ]
            >>> print(chaospy.Pow(2, chaospy.Uniform(-1, 0)).pdf([0.4, 0.6, 0.8, 1.2]))
            [0.         2.40449173 1.8033688  0.        ]
            >>> print(chaospy.Pow(2, 3).pdf([7, 8, 9]))
            [ 0. inf  0.]
        """
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise StochasticallyDependentError(
                    "under-defined distribution {} or {}".format(left, right))

        elif not isinstance(right, Dist):
            return numpy.inf

        else:

            assert numpy.all(left > 0), "imaginary result"
            x_ = numpy.where(xloc <= 0, -numpy.inf,
                    numpy.log(xloc + 1.*(xloc<=0))/numpy.log(left+1.*(left == 1)))
            num_ = numpy.log(left+1.*(left == 1))*xloc
            num_ = num_ + 1.*(num_==0)

            out = evaluation.evaluate_density(right, x_, cache=cache)/num_
            return out

        x_ = numpy.sign(xloc)*numpy.abs(xloc)**(1./right -1)
        xloc = numpy.sign(xloc)*numpy.abs(xloc)**(1./right)
        pairs = numpy.sign(xloc**right) == 1

        out = evaluation.evaluate_density(left, xloc, cache=cache)
        if numpy.any(pairs):
            out = out + pairs*evaluation.evaluate_density(left, -xloc, cache=cache)

        out = numpy.sign(right)*out * x_ / right
        out[numpy.isnan(out)] = numpy.inf

        return out

    def _mom(self, k, left, right, cache):
        """
        Statistical moments.

        Example:
            >>> print(numpy.around(chaospy.Uniform().mom([0, 1, 2, 3]), 4))
            [1.     0.5    0.3333 0.25  ]
            >>> print(numpy.around(chaospy.Pow(chaospy.Uniform(), 2).mom([0, 1, 2, 3]), 4))
            [1.     0.3333 0.2    0.1429]
            >>> print(numpy.around(chaospy.Pow(chaospy.Uniform(1, 2), -1).mom([0, 1, 2, 3]), 4))
            [1.     0.6931 0.5    0.375 ]
            >>> print(numpy.around(chaospy.Pow(2, chaospy.Uniform()).mom([0, 1, 2, 3]), 4))
            [1.     1.4427 2.164  3.3663]
            >>> print(numpy.around(chaospy.Pow(2, chaospy.Uniform(-1, 0)).mom([0, 1, 2, 3]), 4))
            [1.     0.7213 0.541  0.4208]
            >>> print(numpy.around(chaospy.Pow(2, 1).mom([0, 1, 2, 3]), 4))
            [1. 2. 4. 8.]
        """
        if isinstance(right, Dist):
            raise StochasticallyDependentError(
                "distribution as exponent not supported.")
        if not isinstance(left, Dist):
            return left**(right*k)
        if numpy.any(right < 0):
            raise StochasticallyDependentError(
                "distribution to negative power not supported.")
        if not numpy.allclose(right, numpy.array(right, dtype=int)):
            raise StochasticallyDependentError(
                "distribution to fractional power not supported.")
        return evaluation.evaluate_moment(left, k*right, cache=cache)

    def __str__(self):
        """
        Example:
            >>> print(chaospy.Pow(chaospy.Uniform(), 2))
            Pow(Uniform(lower=0, upper=1), 2)
        """
        return (self.__class__.__name__ + "(" + str(self.prm["left"]) +
                ", " + str(self.prm["right"]) + ")")

    def _fwd_cache(self, cache):
        left = evaluation.get_forward_cache(self.prm["left"], cache)
        right = evaluation.get_forward_cache(self.prm["right"], cache)
        if not isinstance(left, Dist) and not isinstance(right, Dist):
            return left**right
        return self

    def _inv_cache(self, cache):
        left = evaluation.get_inverse_cache(self.prm["left"], cache)
        right = evaluation.get_inverse_cache(self.prm["right"], cache)
        if not isinstance(left, Dist) and not isinstance(right, Dist):
            return left**right
        return self



"""Joe copula."""
import numpy
from scipy import special

from .baseclass import Archimedean, Copula
from ..baseclass import Dist

class joe(Archimedean):
    """Joe copula."""

    def __init__(self, length, theta, eps=1e-6):
        assert theta >= 1
        self.length = length
        Dist.__init__(self, th=float(theta), eps=eps)

    def __len__(self):
        return self.length

    def gen(self, x, th):
        return -numpy.log(1-(1-x)**th)

    def igen(self, q, th):
        return 1-(1-numpy.e**-q)**(1/th)


class Joe(Copula):
    """
    Joe Copula

    where `theta` is defined on the interval `[1,inf)`.

    Examples:
        >>> distribution = chaospy.Joe(
        ...     chaospy.Iid(chaospy.Uniform(), 2), theta=2)
        >>> print(distribution)
        Joe(Iid(Uniform(lower=0, upper=1), 2), theta=2)
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> print(numpy.around(distribution.inv(mesh), 4))
        [[[0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]]
        <BLANKLINE>
         [[0.1693 0.2351 0.3964]
          [0.3492 0.4459 0.6346]
          [0.5583 0.6497 0.7949]]]
        >>> print(numpy.around(distribution.fwd(distribution.inv(mesh)), 4))
        [[[0.25 0.5  0.75]
          [0.25 0.5  0.75]
          [0.25 0.5  0.75]]
        <BLANKLINE>
         [[0.25 0.25 0.25]
          [0.5  0.5  0.5 ]
          [0.75 0.75 0.75]]]
        >>> print(numpy.around(distribution.pdf(distribution.inv(mesh)), 4))
        [[1.4454 1.1299 0.8249]
         [1.3193 1.232  1.3345]
         [1.0469 1.1753 1.7359]]
        >>> print(numpy.around(distribution.sample(4), 4))
        [[0.6536 0.115  0.9503 0.4822]
         [0.8222 0.1247 0.3292 0.353 ]]
        >>> print(numpy.around(distribution.mom((1, 2)), 4))
        0.2128
    """

    def __init__(self, dist, theta=2., eps=1e-6):
        """
        Args:
            dist (Dist) : The Distribution to wrap
            theta (float) : Copula parameter
        """
        self._repr = {"theta": theta}
        Copula.__init__(self, dist, joe(len(dist), theta, eps))



"""Skewness operator."""
import numpy

from .. import distributions, poly as polynomials
from .expected import E


def Skew(poly, dist=None, **kws):
    """
    Skewness operator.

    Element by element 3rd order statistics of a distribution or polynomial.

    Args:
        poly (Poly, Dist):
            Input to take skewness on.
        dist (Dist):
            Defines the space the skewness is taken on. It is ignored if
            ``poly`` is a distribution.

    Returns:
        (numpy.ndarray):
            Element for element variance along ``poly``, where
            ``skewness.shape == poly.shape``.

    Examples:
        >>> dist = chaospy.J(chaospy.Gamma(1, 1), chaospy.Normal(0, 2))
        >>> print(chaospy.Skew(dist))
        [2. 0.]
        >>> x, y = chaospy.variable(2)
        >>> poly = chaospy.Poly([1, x, y, 10*x*y])
        >>> print(chaospy.Skew(poly, dist))
        [nan  2.  0.  0.]
    """
    if isinstance(poly, distributions.Dist):
        x = polynomials.variable(len(poly))
        poly, dist = x, poly
    else:
        poly = polynomials.Poly(poly)

    if poly.dim < len(dist):
        polynomials.setdim(poly, len(dist))

    shape = poly.shape
    poly = polynomials.flatten(poly)

    m1 = E(poly, dist)
    m2 = E(poly**2, dist)
    m3 = E(poly**3, dist)
    out = (m3-3*m2*m1+2*m1**3)/(m2-m1**2)**1.5

    out = numpy.reshape(out, shape)
    return out



"""Arc-Tangent Hyperbolic."""
import numpy

from ..baseclass import Dist
from .. import evaluation, approximation



class Arctanh(Dist):
    """
    Arc-Tangent Hyperbolic.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Arctanh(chaospy.Uniform(-0.5, 0.5))
        >>> print(distribution)
        Arctanh(Uniform(lower=-0.5, upper=0.5))
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [-0.3095 -0.1003  0.1003  0.3095]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.91 0.99 0.99 0.91]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 0.1548 -0.4059  0.4851 -0.0178]
        >>> print(numpy.around(distribution.mom(2), 4))
        0.093
    """

    def __init__(self, dist):
        assert isinstance(dist, Dist)
        Dist.__init__(self, dist=dist)

    def _pdf(self, x, dist, cache):
        return evaluation.evaluate_density(
            dist, numpy.tanh(x), cache=cache)/numpy.cosh(x)**2

    def _cdf(self, x, dist, cache):
        return evaluation.evaluate_forward(dist, numpy.tanh(x), cache=cache)

    def _ppf(self, q, dist, cache):
        return numpy.arctanh(evaluation.evaluate_inverse(dist, q, cache=cache))

    def _bnd(self, x, dist, cache):
        return numpy.arctanh(evaluation.evaluate_bound(
            dist, numpy.tanh(x), cache=cache))

    def _mom(self, x, dist, cache):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"

    def _fwd_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.arctanh(dist)
        return self

    def _inv_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.tanh(dist)
        return self



"""
Evaluate probability density function (PDF).

Example usage
-------------

Define a simple distribution and data::

    >>> class Exponential(chaospy.Dist):
    ...     def _pdf(self, x_data, alpha): return alpha*numpy.e**(-alpha*x_data)
    >>> dist = Exponential(alpha=2)
    >>> x_data = numpy.array([[0.1, 0.2, 0.3]])

Normal usage::

    >>> print(numpy.around(evaluate_density(dist, x_data), 4))
    [[1.6375 1.3406 1.0976]]
    >>> print(numpy.around(evaluate_density(dist, x_data, parameters={"alpha": 1}), 4))
    [[0.9048 0.8187 0.7408]]

Cache get triggered if the same distribution is evaluated twice in the same
dependency resolution. The correct behavior for densities is to only return
non-zero values for coordinates that matches. E.g.::

    >>> cache_value = numpy.array([[0.1, 0.2, 0.7]])
    >>> print(numpy.around(evaluate_density(dist, x_data, cache={dist: cache_value}), 4))
    [[1.6375 1.3406 0.    ]]

If a distribution is missing the definition of the density function, it is
instead estimated from cumulative distribution function and boundary function::

    >>> class Exponential(chaospy.Dist):
    ...     def _cdf(self, x_data, alpha): return 1-numpy.e**(-alpha*x_data)
    ...     def _bnd(self, x_data, alpha): return 0, 100
    >>> dist = Exponential(alpha=1)
    >>> print(numpy.around(evaluate_density(dist, x_data), 4))
    [[0.9048 0.8187 0.7408]]
"""
import numpy
from .parameters import load_parameters


def evaluate_density(
        distribution,
        x_data,
        parameters=None,
        cache=None,
):
    """
    Evaluate probability density function (PDF).

    Args:
        distribution (Dist):
            Distribution to evaluate.
        x_data (numpy.ndarray):
            Locations for where evaluate density at.
        parameters (:py:data:typing.Any):
            Collection of parameters to override the default ones in the
            distribution.
        cache (:py:data:typing.Any):
            A collection of previous calculations in case the same distribution
            turns up on more than one occasion.

    Returns:
        The probability density values of ``distribution`` at location
        ``x_data`` using parameters ``parameters``.
    """
    assert len(x_data) == len(distribution)
    assert len(x_data.shape) == 2

    cache = cache if cache is not None else {}
    out = numpy.zeros(x_data.shape)

    # Distribution self know how to handle density evaluation.
    if hasattr(distribution, "_pdf"):
        parameters = load_parameters(
            distribution, "_pdf", parameters=parameters, cache=cache)
        out[:] = distribution._pdf(x_data, **parameters)

    # Approximate density evaluation based on cumulative distribution function.
    else:
        from .. import approximation
        parameters = load_parameters(
            distribution, "_cdf", parameters=parameters, cache=cache)
        out[:] = approximation.approximate_density(
            distribution, x_data, parameters=parameters, cache=cache)

    # dependency handling.
    if distribution in cache:
        out = numpy.where(x_data == cache[distribution], out, 0)
    else:
        cache[distribution] = x_data

    return out



"""
In some cases a constructed distribution that are first and foremost data
driven. In such scenarios it make sense to make use of
`kernel density estimation`_ (KDE). In ``chaospy`` KDE can be accessed through
the :func:`SampleDist` constructor.

Basic usage of the :func:`SampleDist` constructor involves just passing the
data as input argument::

    >>> data = [3, 4, 5, 5]
    >>> distribution = chaospy.SampleDist(data)

This distribution can be used as any other distributions::

    >>> print(numpy.around(distribution.cdf([3, 3.5, 4, 4.5, 5]), 4))
    [0.     0.1932 0.4279 0.7043 1.    ]
    >>> print(numpy.around(distribution.mom(1), 4))
    4.0922

It also supports lower and upper bounds defining where the range is expected to
appear, which gives a slightly different distribution::

    >>> distribution = chaospy.SampleDist(data, lo=2, up=6)
    >>> print(numpy.around(distribution.cdf([3, 3.5, 4, 4.5, 5]), 4))
    [0.1344 0.2543 0.4001 0.5716 0.7552]
    >>> print(numpy.around(distribution.mom(1), 4))
    4.2149

Currently the wrapper is limited to only support univariate distributions.

.. _kernel density estimation: \
https://en.wikipedia.org/wiki/Kernel_density_estimation
"""
import numpy
from scipy.stats import gaussian_kde

from chaospy.distributions.baseclass import Dist
from chaospy.distributions.operators.addition import Add
from chaospy.distributions.collection.uniform import Uniform
from chaospy.distributions.collection.deprecate import deprecation_warning


class sample_dist(Dist):
    """A distribution that is based on a kernel density estimator (KDE)."""
    def __init__(self, samples, lo, up):
        self.samples = samples
        self.kernel = gaussian_kde(samples, bw_method="scott")
        self.flo = self.kernel.integrate_box_1d(0, lo)
        self.fup = self.kernel.integrate_box_1d(0, up)
        super(sample_dist, self).__init__(lo=lo, up=up)

    def _cdf(self, x, lo, up):
        cdf_vals = numpy.zeros(x.shape)
        for i in range(0, len(x)):
            cdf_vals[i] = [self.kernel.integrate_box_1d(0, x_i) for x_i in x[i]]
        cdf_vals = (cdf_vals - self.flo) / (self.fup - self.flo)
        return cdf_vals

    def _pdf(self, x, lo, up):
        return self.kernel(x)

    def _bnd(self, x, lo, up):
        return (lo, up)

    def sample(self, size=(), rule="R", antithetic=None, verbose=False, **kws):
        """
        Overwrite sample() function, because the constructed Dist that is
        based on the KDE is only working with the random sampling that is
        given by the KDE itself.
        """
        size_ = numpy.prod(size, dtype=int)
        dim = len(self)
        if dim > 1:
            if isinstance(size, (tuple,list,numpy.ndarray)):
                shape = (dim,) + tuple(size)
            else:
                shape = (dim, size)
        else:
            shape = size

        out = self.kernel.resample(size_)[0]
        try:
            out = out.reshape(shape)
        except:
            if len(self) == 1:
                out = out.flatten()
            else:
                out = out.reshape(dim, out.size/dim)

        return out


def SampleDist(samples, lo=None, up=None):
    """
    Distribution based on samples.

    Estimates a distribution from the given samples by constructing a kernel
    density estimator (KDE).

    Args:
        samples (numpy.ndarray):
            Sample values to construction of the KDE
        lo (float):
            Location of lower threshold
        up (float):
            Location of upper threshold

    Example:
        >>> distribution = chaospy.SampleDist([0, 1, 1, 1, 2])
        >>> print(distribution)
        sample_dist(lo=0, up=2)
        >>> q = numpy.linspace(0, 1, 5)
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.     0.6016 1.     1.3984 2.    ]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.   0.25 0.5  0.75 1.  ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.2254 0.4272 0.5135 0.4272 0.2254]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 1.5877  1.1645 -0.0131  1.3302]
        >>> print(numpy.around(distribution.mom(1), 4))
        1.0
    """
    samples = numpy.asarray(samples)
    if lo is None:
        lo = samples.min()
    if up is None:
        up = samples.max()

    try:
        #construct the kernel density estimator
        dist = sample_dist(samples, lo, up)

    #raised by gaussian_kde if dataset is singular matrix
    except numpy.linalg.LinAlgError:
        dist = Uniform(lower=-numpy.inf, upper=numpy.inf)

    return dist



"""
Addition operator.

Example usage
-------------

Distribution and a constant::

    >>> distribution = chaospy.Normal(0, 1) + 10
    >>> print(distribution)
    Add(Normal(mu=0, sigma=1), 10)
    >>> print(numpy.around(distribution.sample(5), 4))
    [10.395   8.7997 11.6476  9.9553 11.1382]
    >>> print(numpy.around(distribution.fwd([9, 10, 11]), 4))
    [0.1587 0.5    0.8413]
    >>> print(numpy.around(distribution.inv(distribution.fwd([9, 10, 11])), 4))
    [ 9. 10. 11.]
    >>> print(numpy.around(distribution.pdf([9, 10, 11]), 4))
    [0.242  0.3989 0.242 ]
    >>> print(distribution.mom([1, 2, 3]))
    [  10.  101. 1030.]
    >>> print(distribution.ttr([1, 2, 3]))
    [[10. 10. 10.]
     [ 1.  2.  3.]]

Construct joint addition distribution::

    >>> lhs = chaospy.Uniform(2, 3)
    >>> rhs = chaospy.Uniform(3, 4)
    >>> addition = lhs + rhs
    >>> print(addition)
    Add(Uniform(lower=2, upper=3), Uniform(lower=3, upper=4))
    >>> joint1 = chaospy.J(lhs, addition)
    >>> joint2 = chaospy.J(rhs, addition)

Generate random samples::

    >>> print(numpy.around(joint1.sample(4), 4))
    [[2.2123 2.0407 2.3972 2.2331]
     [6.0541 5.2478 6.1397 5.6253]]
    >>> print(numpy.around(joint2.sample(4), 4))
    [[3.1823 3.7435 3.0696 3.8853]
     [6.1349 6.6747 5.485  5.9143]]

Forward transformations::

    >>> lcorr = numpy.array([2.1, 2.5, 2.9])
    >>> rcorr = numpy.array([3.01, 3.5, 3.99])
    >>> print(numpy.around(joint1.fwd([lcorr, lcorr+rcorr]), 4))
    [[0.1  0.5  0.9 ]
     [0.01 0.5  0.99]]
    >>> print(numpy.around(joint2.fwd([rcorr, lcorr+rcorr]), 4))
    [[0.01 0.5  0.99]
     [0.1  0.5  0.9 ]]

Inverse transformations::

    >>> print(numpy.around(joint1.inv(joint1.fwd([lcorr, lcorr+rcorr])), 4))
    [[2.1  2.5  2.9 ]
     [5.11 6.   6.89]]
    >>> print(numpy.around(joint2.inv(joint2.fwd([rcorr, lcorr+rcorr])), 4))
    [[3.01 3.5  3.99]
     [5.11 6.   6.89]]

Raw moments::

    >>> print(numpy.around(joint1.mom([(0, 1, 1), (1, 0, 1)]), 4))
    [ 6.   2.5 15. ]
    >>> print(numpy.around(joint2.mom([(0, 1, 1), (1, 0, 1)]), 4))
    [ 6.   3.5 21. ]
"""
from __future__ import division
from scipy.special import comb
import numpy

from ..baseclass import Dist, StochasticallyDependentError
from .. import evaluation


class Add(Dist):
    """Addition."""

    def __init__(self, left, right):
        """
        Args:
            left (Dist, numpy.ndarray) : Left hand side.
            right (Dist, numpy.ndarray) : Right hand side.
        """
        Dist.__init__(self, left=left, right=right)

    def _bnd(self, xloc, left, right, cache):
        """
        Distribution bounds.

        Example:
            >>> print(chaospy.Uniform().range([-2, 0, 2, 4]))
            [[0. 0. 0. 0.]
             [1. 1. 1. 1.]]
            >>> print(chaospy.Add(chaospy.Uniform(), 2).range([-2, 0, 2, 4]))
            [[2. 2. 2. 2.]
             [3. 3. 3. 3.]]
            >>> print(chaospy.Add(2, chaospy.Uniform()).range([-2, 0, 2, 4]))
            [[2. 2. 2. 2.]
             [3. 3. 3. 3.]]
            >>> print(chaospy.Add(1, 1).range([-2, 0, 2, 4]))
            [[2. 2. 2. 2.]
             [2. 2. 2. 2.]]
        """
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))
        elif not isinstance(right, Dist):
            return left+right, left+right
        else:
            left, right = right, left

        right = numpy.asfarray(right)
        if len(right.shape) == 3:
            xloc_ = (xloc.T-right[0].T).T
            lower, upper = evaluation.evaluate_bound(left, xloc_, cache=cache.copy())
            lower0, upper0 = (lower.T+right[0].T).T, (upper.T+right[0].T).T

            xloc_ = (xloc.T-right[1].T).T
            lower, upper = evaluation.evaluate_bound(left, xloc_, cache=cache)
            lower1, upper1 = (lower.T+right[1].T).T, (upper.T+right[1].T).T

            lower = numpy.min([lower0, lower1], 0)
            upper = numpy.max([upper0, upper1], 0)

        else:
            xloc_ = (xloc.T-right.T).T
            lower, upper = evaluation.evaluate_bound(left, xloc_, cache=cache.copy())
            lower, upper = (lower.T+right.T).T, (upper.T+right.T).T

        assert lower.shape == xloc.shape
        assert upper.shape == xloc.shape
        return lower, upper

    def _cdf(self, xloc, left, right, cache):
        """
        Cumulative distribution function.

        Example:
            >>> print(chaospy.Uniform().fwd([-0.5, 0.5, 1.5, 2.5]))
            [0.  0.5 1.  1. ]
            >>> print(chaospy.Add(chaospy.Uniform(), 1).fwd([-0.5, 0.5, 1.5, 2.5]))
            [0.  0.  0.5 1. ]
            >>> print(chaospy.Add(1, chaospy.Uniform()).fwd([-0.5, 0.5, 1.5, 2.5]))
            [0.  0.  0.5 1. ]
            >>> print(chaospy.Add(1, 1).fwd([-0.5, 0.5, 1.5, 2.5]))
            [0. 0. 0. 1.]
        """
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))
        elif not isinstance(right, Dist):
            return numpy.asfarray(left+right <= xloc)
        else:
            left, right = right, left
        xloc = (xloc.T-numpy.asfarray(right).T).T
        output = evaluation.evaluate_forward(left, xloc, cache=cache)
        assert output.shape == xloc.shape
        return output

    def _pdf(self, xloc, left, right, cache):
        """
        Probability density function.

        Example:
            >>> print(chaospy.Uniform().pdf([-2, 0, 2, 4]))
            [0. 1. 0. 0.]
            >>> print(chaospy.Add(chaospy.Uniform(), 2).pdf([-2, 0, 2, 4]))
            [0. 0. 1. 0.]
            >>> print(chaospy.Add(2, chaospy.Uniform()).pdf([-2, 0, 2, 4]))
            [0. 0. 1. 0.]
            >>> print(chaospy.Add(1, 1).pdf([-2, 0, 2, 4])) # Dirac logic
            [ 0.  0. inf  0.]
        """
        left = evaluation.get_forward_cache(left, cache)
        right = evaluation.get_forward_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))
        elif not isinstance(right, Dist):
            return numpy.inf
        else:
            left, right = right, left

        xloc = (xloc.T-numpy.asfarray(right).T).T
        output = evaluation.evaluate_density(left, xloc, cache=cache)
        assert output.shape == xloc.shape
        return output

    def _ppf(self, uloc, left, right, cache):
        """
        Point percentile function.

        Example:
            >>> print(chaospy.Uniform().inv([0.1, 0.2, 0.9]))
            [0.1 0.2 0.9]
            >>> print(chaospy.Add(chaospy.Uniform(), 2).inv([0.1, 0.2, 0.9]))
            [2.1 2.2 2.9]
            >>> print(chaospy.Add(2, chaospy.Uniform()).inv([0.1, 0.2, 0.9]))
            [2.1 2.2 2.9]
            >>> print(chaospy.Add(1, 1).inv([0.1, 0.2, 0.9]))
            [2. 2. 2.]
        """
        left = evaluation.get_inverse_cache(left, cache)
        right = evaluation.get_inverse_cache(right, cache)

        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise evaluation.DependencyError(
                    "under-defined distribution {} or {}".format(left, right))
        elif not isinstance(right, Dist):
            return left+right
        else:
            left, right = right, left

        xloc = evaluation.evaluate_inverse(left, uloc, cache=cache)
        output = (xloc.T + numpy.asfarray(right).T).T
        return output

    def _mom(self, keys, left, right, cache):
        """
        Statistical moments.

        Example:
            >>> print(numpy.around(chaospy.Uniform().mom([0, 1, 2, 3]), 4))
            [1.     0.5    0.3333 0.25  ]
            >>> print(numpy.around(chaospy.Add(chaospy.Uniform(), 2).mom([0, 1, 2, 3]), 4))
            [ 1.      2.5     6.3333 16.25  ]
            >>> print(numpy.around(chaospy.Add(2, chaospy.Uniform()).mom([0, 1, 2, 3]), 4))
            [ 1.      2.5     6.3333 16.25  ]
            >>> print(numpy.around(chaospy.Add(1, 1).mom([0, 1, 2, 3]), 4))
            [1. 2. 4. 8.]
        """
        if evaluation.get_dependencies(left, right):
            raise evaluation.DependencyError(
                "sum of dependent distributions not feasible: "
                "{} and {}".format(left, right)
            )

        keys_ = numpy.mgrid[tuple(slice(0, key+1, 1) for key in keys)]
        keys_ = keys_.reshape(len(self), -1)

        if isinstance(left, Dist):
            left = [
                evaluation.evaluate_moment(left, key, cache=cache)
                for key in keys_.T
            ]
        else:
            left = list(reversed(numpy.array(left).T**keys_.T))
        if isinstance(right, Dist):
            right = [
                evaluation.evaluate_moment(right, key, cache=cache)
                for key in keys_.T
            ]
        else:
            right = list(reversed(numpy.array(right).T**keys_.T))

        out = numpy.zeros(keys.shape)
        for idx in range(keys_.shape[1]):
            key = keys_.T[idx]
            coef = comb(keys.T, key)
            out += coef*left[idx]*right[idx]*(key <= keys.T)

        if len(self) > 1:
            out = numpy.prod(out, 1)
        return out

    def _ttr(self, kloc, left, right, cache):
        """
        Three terms recursion coefficients.

        Example:
            >>> print(numpy.around(chaospy.Uniform().ttr([0, 1, 2, 3]), 4))
            [[ 0.5     0.5     0.5     0.5   ]
             [-0.      0.0833  0.0667  0.0643]]
            >>> print(numpy.around(chaospy.Add(chaospy.Uniform(), 2).ttr([0, 1, 2, 3]), 4))
            [[ 2.5     2.5     2.5     2.5   ]
             [-0.      0.0833  0.0667  0.0643]]
            >>> print(numpy.around(chaospy.Add(2, chaospy.Uniform()).ttr([0, 1, 2, 3]), 4))
            [[ 2.5     2.5     2.5     2.5   ]
             [-0.      0.0833  0.0667  0.0643]]
            >>> print(numpy.around(chaospy.Add(1, 1).ttr([0, 1, 2, 3]), 4)) # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
                ...
            chaospy.distributions.baseclass.StochasticallyDependentError: recurrence ...
        """
        if isinstance(left, Dist):
            if isinstance(right, Dist):
                raise StochasticallyDependentError(
                    "sum of distributions not feasible: "
                    "{} and {}".format(left, right)
                )
        else:
            if not isinstance(right, Dist):
                raise StochasticallyDependentError(
                    "recurrence coefficients for constants not feasible: "
                    "{}".format(left+right)
                )
            left, right = right, left

        coeff0, coeff1 = evaluation.evaluate_recurrence_coefficients(
            left, kloc, cache=cache)
        return coeff0 + numpy.asarray(right), coeff1

    def __str__(self):
        if self._repr is None:
            return (self.__class__.__name__ + "(" + str(self.prm["left"]) +
                    ", " + str(self.prm["right"]) + ")")
        return super(Add, self).__str__()

    def _fwd_cache(self, cache):
        left = evaluation.get_forward_cache(self.prm["left"], cache)
        right = evaluation.get_forward_cache(self.prm["right"], cache)
        if not isinstance(left, Dist) and not isinstance(right, Dist):
            return left+right
        return self

    def _inv_cache(self, cache):
        left = evaluation.get_forward_cache(self.prm["left"], cache)
        right = evaluation.get_forward_cache(self.prm["right"], cache)
        if not isinstance(left, Dist) and not isinstance(right, Dist):
            return left+right
        return self


def add(left, right):
    return Add(left, right)



"""
Negative of a distribution.

Example usage
-------------

Invert sign of a distribution::

    >>> distribution = -chaospy.Uniform(0, 1)
    >>> print(distribution)
    Neg(Uniform(lower=0, upper=1))
    >>> print(numpy.around(distribution.sample(5), 4))
    [-0.3464 -0.885  -0.0497 -0.5178 -0.1275]
    >>> print(distribution.fwd([-0.3, -0.2, -0.1]))
    [0.7 0.8 0.9]
    >>> print(distribution.inv(distribution.fwd([-0.3, -0.2, -0.1])))
    [-0.3 -0.2 -0.1]
    >>> print(distribution.pdf([-0.3, -0.2, -0.1]))
    [1. 1. 1.]
    >>> print(numpy.around(distribution.mom([1, 2, 3]), 4))
    [-0.5     0.3333 -0.25  ]
    >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
    [[-0.5    -0.5    -0.5   ]
     [ 0.0833  0.0667  0.0643]]

"""
import numpy
from ..baseclass import Dist
from .. import evaluation


class Neg(Dist):
    """Negative of a distribution."""

    def __init__(self, dist):
        """
        Constructor.

        Args:
            dist (Dist) : distribution.
        """
        Dist.__init__(self, dist=dist)

    def _bnd(self, xloc, dist, cache):
        """Distribution bounds."""
        return -evaluation.evaluate_bound(dist, -xloc)[::-1]

    def _pdf(self, xloc, dist, cache):
        """Probability density function."""
        return evaluation.evaluate_density(dist, -xloc, cache=cache)

    def _cdf(self, xloc, dist, cache):
        """Cumulative distribution function."""
        return 1-evaluation.evaluate_forward(dist, -xloc, cache=cache)

    def _ppf(self, q, dist, cache):
        """Point percentile function."""
        return -evaluation.evaluate_inverse(dist, 1-q, cache=cache)

    def _mom(self, k, dist, cache):
        """Statistical moments."""
        return (-1)**numpy.sum(k)*evaluation.evaluate_moment(
            dist, k, cache=cache)

    def _ttr(self, k, dist, cache):
        """Three terms recursion coefficients."""
        a,b = evaluation.evaluate_recurrence_coefficients(dist, k)
        return -a, b

    def __str__(self):
        """String representation."""
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"

    def _fwd_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return -dist
        return self

    def _inv_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return -dist
        return self


def neg(left):
    """
    Negative of a distribution.

    Args:
        dist (Dist) : distribution.
    """
    return Neg(left)



"""
Hermite Genz-Keister quadrature rules

Adapted from John Burkardt's implementation in Matlab
"""
from .frontend import quad_genz_keister, GENS_KEISTER_FUNCTIONS

from .gk16 import quad_genz_keister_16
from .gk18 import quad_genz_keister_18
from .gk22 import quad_genz_keister_22
from .gk24 import quad_genz_keister_24



"""Triangle probability distribution."""
import numpy
from scipy import special, misc

from ..baseclass import Dist
from ..operators.addition import Add
from .beta import beta_


class triangle(Dist):
    """Triangle probability distribution."""

    def __init__(self, a=.5):
        assert numpy.all(a>=0) and numpy.all(a<=1)
        Dist.__init__(self, a=a)

    def _pdf(self, D, a):
        return numpy.where(D<a, 2*D/a, 2*(1-D)/(1-a))

    def _cdf(self, D, a):
        return numpy.where(D<a, D**2/(a + (a==0)),
                (2*D-D*D-a)/(1-a+(a==1)))

    def _ppf(self, q, a):
        return numpy.where(q<a, numpy.sqrt(q*a), 1-numpy.sqrt(1-a-q*(1-a)))

    def _mom(self, k, a):
        a_ = a*(a!=1)
        out = 2*(1.-a_**(k+1))/((k+1)*(k+2)*(1-a_))
        return numpy.where(a==1, 2./(k+2), out)

    def _bnd(self, x, a):
        return 0., 1.

    def _ttr(self, k, a):
        if a == 0:
            return beta_()._ttr(k, 1, 2)
        if a == 1:
            return beta_()._ttr(k, 2, 1)

        from ...quadrature import quad_fejer, discretized_stieltjes
        q1, w1 = quad_fejer(int(1000*a), (0, a))
        q2, w2 = quad_fejer(int(1000*(1-a)), (a, 1))
        q = numpy.concatenate([q1,q2], 1)
        w = numpy.concatenate([w1,w2])*self.pdf(q[0])

        coeffs, _, _ = discretized_stieltjes(k, q, w)
        return coeffs[:, 0, -1]


class Triangle(Add):
    """
    Triangle Distribution.

    Must have lower <= midpoint <= upper.

    Args:
        lower (float, Dist) : Lower bound
        midpoint (float, Dist) : Location of the top
        upper (float, Dist) : Upper bound

    Examples:
        >>> distribution = chaospy.Triangle(2, 3, 4)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.6325 2.8944 3.1056 3.3675]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.6325 0.8944 0.8944 0.6325]
        >>> print(numpy.around(distribution.sample(4), 4))
        [3.1676 2.4796 3.6847 2.982 ]
        >>> print(numpy.around(distribution.mom(1), 4))
        3.0
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[3.     3.     3.    ]
         [0.1667 0.2333 0.2327]]
    """

    def __init__(self, lower=-1, midpoint=0, upper=1):
        self._repr = {"lower": lower, "midpoint": midpoint, "upper": upper}
        midpoint = (midpoint-lower)*1./(upper-lower)
        Add.__init__(self, left=triangle(midpoint)*(upper-lower), right=lower)



"""Variance based sensitivity decomposition using Sobol Indices."""
from .main import Sens_m
from .main2 import Sens_m2
from .total import Sens_t
from .nataf import Sens_m_nataf, Sens_t_nataf, Sens_nataf



"""Weibull Distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class weibull(Dist):
    """Weibull Distribution."""

    def __init__(self, a=1):
        Dist.__init__(self, a=a)

    def _pdf(self, x, a):
        return a*x**(a-1)*numpy.e**(-x**a)

    def _cdf(self, x, a):
        return (1-numpy.e**(-x**a))

    def _ppf(self, q, a):
        return (-numpy.log(1-q+1*(q==1)))**(1./a)*(q!=1) + 30.**(1./a)*(q==1)

    def _mom(self, k, a):
        return special.gamma(1.+k*1./a)

    def _bnd(self, x, a):
        return 0, 30.**(1./a)


class Weibull(Add):
    """
    Weibull Distribution

    Args:
        shape (float, Dist) : Shape parameter.
        scale (float, Dist) : Scale parameter.
        shift (float, Dist) : Location of lower bound.

    Examples:
        >>> distribution = chaospy.Weibull(2)
        >>> print(distribution)
        Weibull(scale=1, shape=2, shift=0)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.4724 0.7147 0.9572 1.2686]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.7558 0.8577 0.7658 0.5075]
        >>> print(numpy.around(distribution.sample(4), 4))
        [1.0296 0.3495 1.7325 0.8113]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.8862
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[1.1786 1.4264 1.6407]
         [0.2146 0.3963 0.5691]]
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=weibull(shape)*scale, right=shift)



r"""
Creating an orthogonal polynomial expansion can be numerical unstable
when using raw statistical moments as input (see `paper by Gautschi`_ for
details). This can be a problem if constructing large expansions since the
error blows up. Given that the distribution is univariate it is instead
possible to create orthogonal polynomials stabilized using the three terms
recursion relation:

.. math::
    \Phi_{n+1}(q) = \Phi_{n}(q) (q-A_n) - \Phi_{n-1}(q) B_n,

where

.. math::
    A_n = \frac{\left\langle q \Phi_n, \Phi_n \right\rangle}{
          \left\langle \Phi_n, \Phi_n \right\rangle}
        = \frac{\mathbb E[q\Phi_{n}^2]}{
          \mathbb E[\Phi_n^2]}
    B_n = \frac{\left\langle \Phi_n, \Phi_n \right\rangle}{
          \left\langle \Phi_{n-1}, \Phi_{n-1} \right\rangle}
        = \frac{\mathbb E[\Phi_{n}^2]}{\mathbb E[\Phi_{n-1}^2]}

A multivariate polynomial expansion can be created using tensor product
rule of univariate polynomials expansions. This assumes that the
distribution is stochastically independent.

In the ``chaospy`` toolbox three terms recursion coefficient can be
generating by calling the ``ttr`` instance method::

    >>> dist = chaospy.Uniform(-1,1)
    >>> print(numpy.around(dist.ttr([0,1,2,3]), 4))
    [[ 0.      0.      0.      0.    ]
     [-0.      0.3333  0.2667  0.2571]]

In many of the pre-defined probability distributions in ``chaospy``, the three
terms recursion coefficients are calculated analytically. If the distribution
does not support the method, the coefficients are instead calculated using the
discretized Stieltjes method (described in the `paper by Golub and Welsch`_).

In ``chaospy`` constructing orthogonal polynomial using the three term
recursion scheme can be done through ``orth_ttr``. For example::

    >>> dist = chaospy.Iid(chaospy.Gamma(1), 2)
    >>> orths = chaospy.orth_ttr(2, dist)
    >>> print(chaospy.around(orths, 4))
    [1.0, q1-1.0, q0-1.0, q1^2-4.0q1+2.0, q0q1-q0-q1+1.0, q0^2-4.0q0+2.0]

The method will use the ``ttr`` function if available, and discretized
Stieltjes otherwise.

.. _paper by Gautschi: https://www.ams.org/journals/mcom/1968-22-102/S0025-5718-1968-0228171-0/
.. _paper by Golub and Welsch: https://web.stanford.edu/class/cme335/spr11/S0025-5718-69-99647-1.pdf
"""
import numpy
import chaospy


def orth_ttr(
        order, dist, normed=False, sort="G", retall=False,
        cross_truncation=1., **kws):
    """
    Create orthogonal polynomial expansion from three terms recursion formula.

    Args:
        order (int):
            Order of polynomial expansion.
        dist (Dist):
            Distribution space where polynomials are orthogonal If dist.ttr
            exists, it will be used. Must be stochastically independent.
        normed (bool):
            If True orthonormal polynomials will be used.
        sort (str):
            Polynomial sorting. Same as in basis.
        retall (bool):
            If true return numerical stabilized norms as well. Roughly the same
            as ``cp.E(orth**2, dist)``.
        cross_truncation (float):
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion. only include terms where the exponents ``K``
            satisfied the equation
            ``order >= sum(K**(1/cross_truncation))**cross_truncation``.

    Returns:
        (Poly, numpy.ndarray):
            Orthogonal polynomial expansion and norms of the orthogonal
            expansion on the form ``E(orth**2, dist)``. Calculated using
            recurrence coefficients for stability.

    Examples:
        >>> Z = chaospy.Normal()
        >>> print(chaospy.around(chaospy.orth_ttr(4, Z), 4))
        [1.0, q0, q0^2-1.0, q0^3-3.0q0, q0^4-6.0q0^2+3.0]
    """
    try:
        _, polynomials, norms, = chaospy.quadrature.recurrence.analytical_stieljes(
            numpy.max(order), dist, normed=normed)
    except NotImplementedError:
        abscissas, weights = chaospy.quadrature.generate_quadrature(
            int(10000**(1/len(dist))), dist, rule="F")
        _, polynomials, norms, = chaospy.quadrature.recurrence.discretized_stieltjes(
            numpy.max(order), abscissas, weights, normed=normed)

    polynomials = chaospy.poly.reshape(
        polynomials, (len(dist), numpy.max(order)+1))

    indices = chaospy.bertran.bindex(
        start=0, stop=order, dim=len(dist), sort=sort,
        cross_truncation=cross_truncation,
    )
    if len(dist) > 1:
        polynomials = chaospy.poly.prod(chaospy.poly.Poly([
            poly[idx] for poly, idx in zip(polynomials, indices.T)]), 0)
        norms = numpy.prod([
            norms_[idx] for norms_, idx in zip(norms, indices.T)], 0)
    else:
        polynomials = chaospy.poly.flatten(polynomials)

    if retall:
        return polynomials, norms
    return polynomials



"""Truncated normal distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from .deprecate import deprecation_warning


class TruncNormal(Dist):
    """
    Truncated normal distribution

    Args:
        lower (float, Dist): Location of lower threshold
        upper (float, Dist): Location of upper threshold
        mu (float, Dist): Mean of normal distribution
        sigma (float, Dist): Standard deviation of normal distribution

    Examples:
        >>> distribution = chaospy.TruncNormal(2, 4, 2, 2)
        >>> print(distribution)
        TruncNormal(lower=2, mu=2, sigma=2, upper=4)
        >>> q = numpy.linspace(0, 1, 5)
        >>> print(numpy.around(distribution.inv(q), 4))
        [2.     2.4311 2.8835 3.387  4.    ]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.   0.25 0.5  0.75 1.  ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [1.1687 1.1419 1.0601 0.9189 0.7089]
        >>> print(numpy.around(distribution.sample(4), 4))
        [3.1841 2.1971 3.8643 2.8501]
    """

    def __init__(self, lower=-1, upper=1, mu=0, sigma=1):
        self._repr = {"lower": lower, "upper": upper, "mu": mu, "sigma": sigma}
        Dist.__init__(self, a=lower, b=upper, sigma=sigma, mu=mu)

    def _pdf(self, x, a, b, mu, sigma):
        fa = special.ndtr((a-mu)/sigma)
        fb = special.ndtr((b-mu)/sigma)
        x = (x-mu)/sigma
        norm = (2*numpy.pi)**(-.5)*numpy.e**(-x**2/2.)
        return norm/(fb-fa)

    def _cdf(self, x, a, b, mu, sigma):
        fa = special.ndtr((a-mu)/sigma)
        fb = special.ndtr((b-mu)/sigma)
        x = special.ndtr((x-mu)/sigma)
        return (x - fa) / (fb-fa)

    def _ppf(self, q, a, b, mu, sigma):
        fa = special.ndtr((a-mu)/sigma)
        fb = special.ndtr((b-mu)/sigma)
        return special.ndtri(q*(fb-fa) + fa)*sigma + mu

    def _bnd(self, x, a, b, mu, sigma):
        return a, b


Truncnorm = deprecation_warning(TruncNormal, "Truncnorm")



"""Non-central Chi-squared distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class chi_squared(Dist):
    """Central Chi-squared distribution."""

    def __init__(self, df, nc):
        Dist.__init__(self, df=df, nc=nc)

    def _pdf(self, x, df, nc):
        output = 0.5*numpy.e**(-0.5*(x+nc))
        output *= (x/nc)**(0.25*df-0.5)
        output *= special.iv(0.5*df-1, (nc*x)**0.5)
        return output

    def _cdf(self, x, df, nc):
        return special.chndtr(x, df, nc)

    def _ppf(self, q, df, nc):
        return special.chndtrix(q, df, nc)

    def _bnd(self, x, df, nc):
        for expon in range(12, -1, -1):
            upper = self._ppf(1-10**-expon, df, nc)
            if not numpy.isnan(upper).item():
                return 0., upper


class ChiSquared(Add):
    """
    (Non-central) Chi-squared distribution.

    Examples:
        >>> distribution = chaospy.ChiSquared(2, 4, 1, 1)
        >>> print(distribution)
        ChiSquared(df=2, nc=1, scale=4, shift=1)
        >>> q = numpy.linspace(0, 1, 7)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [ 3.369   6.1849  9.7082 14.5166 22.4295]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.1667 0.3333 0.5    0.6667 0.8333]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.065  0.0536 0.0414 0.0286 0.0149]
        >>> print(numpy.around(distribution.sample(4), 4))
        [14.0669  2.595  35.6294  9.2851]
        >>> print(numpy.around(distribution.mom(1), 4))
        13.0001
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[ 33.      51.6666  69.8034]
         [128.     432.     895.9962]]
    """

    def __init__(self, df=1, scale=1, shift=0, nc=0):
        """
        Args:
            df (float, Dist) : Degrees of freedom
            scale (float, Dist) : Scaling parameter
            shift (float, Dist) : Location parameter
            nc (float, Dist) : Non-centrality parameter
        """
        self._repr = {"df": df, "scale": scale, "shift": shift, "nc": nc}
        Add.__init__(self, left=chi_squared(df, nc)*scale, right=shift)


Chisquared = deprecation_warning(ChiSquared, "Chisquared")



"""Double Weibull distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class double_weibull(Dist):
    """Double weibull distribution."""

    def __init__(self, c):
        Dist.__init__(self, c=c)

    def _pdf(self, x, c):
        ax = numpy.abs(x)
        Px = c/2.0*ax**(c-1.0)*numpy.exp(-ax**c)
        return Px

    def _cdf(self, x, c):
        Cx1 = 0.5*numpy.exp(-abs(x)**c)
        return numpy.where(x > 0, 1-Cx1, Cx1)

    def _ppf(self, q, c):
        q_ = numpy.where(q>.5, 1-q, q)
        c = c*numpy.ones(q_.shape)
        Cq1 = numpy.ones(q_.shape)
        indices = q_ == 0
        Cq1[indices] = 1e10
        indices = ~indices & (c != 0)
        Cq1[indices] = (-numpy.log(2*q_[indices]))**(1./c[indices])
        return numpy.where(q>.5, Cq1, -Cq1)

    def _bnd(self, x, c):
        return self._ppf(1e-10, c), self._ppf(1-1e-10, c)


class DoubleWeibull(Add):
    """
    Double Weibull distribution.

    Args:
        shape (float, Dist): Shape parameter
        scale (float, Dist): Scaling parameter
        shift (float, Dist): Location parameter

    Examples:
        >>> distribution = chaospy.DoubleWeibull(2, 4, 2)
        >>> print(distribution)
        DoubleWeibull(scale=4, shape=2, shift=2)
        >>> q = numpy.linspace(0, 1, 5)
        >>> print(numpy.around(distribution.inv(q), 4))
        [-16.903   -1.3302   2.       5.3302  20.903 ]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.   0.25 0.5  0.75 1.  ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.     0.1041 0.     0.1041 0.    ]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 4.4232 -2.8491  8.0772  1.2382]
        >>> print(numpy.around(distribution.mom(1), 4))
        2.0
    """

    def __init__(self, shape=1, scale=1, shift=0):
        self._repr = {"shape": shape, "scale": scale, "shift": shift}
        Add.__init__(self, left=double_weibull(shape)*scale, right=shift)


Dbl_weibull = deprecation_warning(DoubleWeibull, "Dbl_weibull")



"""Discrete uniform probability distribution."""
import numpy

from ..baseclass import Dist


class DiscreteUniform(Dist):
    """
    Discrete uniform probability distribution.

    Args:
        lower (float, chaospy.Dist):
            Lower threshold of distribution. Must be smaller than ``upper``.
            Value will be rounded up to closes integer.
        upper (float, chaospy.Dist):
            Upper threshold of distribution. Value will be rouned down to
            closes integer.

    Examples:
        >>> distribution = chaospy.DiscreteUniform(2, 4)
        >>> print(distribution)
        DiscreteUniform(lower=2, upper=4)
        >>> q = numpy.linspace(0, 1, 9)
        >>> print(numpy.around(distribution.inv(q), 4))
        [2 2 2 3 3 3 4 4 4]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.  0.  0.  0.5 0.5 0.5 1.  1.  1. ]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5]
        >>> print(numpy.around(distribution.sample(4), 4))
        [3 2 4 3]
        >>> print(numpy.around(distribution.mom(1), 4))
        3.0
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[3.     3.     3.3333]
         [0.6667 0.3333 0.    ]]
    """
    interpret_as_integer = True

    def __init__(self, lower, upper):
        Dist.__init__(self, lower=lower, upper=upper)

    def _cdf(self, x_data, lower, upper):
        """Cumulative distribution function."""
        return ((numpy.floor(x_data)-numpy.ceil(lower))/
                (numpy.floor(upper)-numpy.ceil(lower)))

    def _bnd(self, x_data, lower, upper):
        """Lower and upper bounds."""
        return numpy.ceil(lower), numpy.floor(upper)

    def _pdf(self, x_data, lower, upper):
        """Probability density function."""
        return x_data**0/(numpy.floor(upper)-numpy.ceil(lower))

    def _ppf(self, q_data, lower, upper):
        """Point percentile function."""
        return (numpy.floor(q_data*(
            numpy.floor(upper+1)-numpy.ceil(lower))+numpy.ceil(lower)))

    def _mom(self, k_data, lower, upper):
        """Raw statistical moments."""
        return numpy.mean(numpy.arange(
            numpy.ceil(lower), numpy.floor(upper)+1)**k_data)

    def _ttr(self, k_data, lower, upper):
        """Three terms recurrence coefficients."""
        from chaospy.quadrature import discretized_stieltjes
        abscissas = numpy.arange(numpy.ceil(lower), numpy.floor(upper)+1)
        weights = numpy.repeat(1./len(abscissas), len(abscissas))
        (alpha, beta), _, _ = discretized_stieltjes(k_data, [abscissas], weights)
        return alpha[0, -1], beta[0, -1]



"""
Basic tools for Bertran index manipulation.
"""
import functools
import itertools

import numpy
from scipy.special import comb

import chaospy.bertran
from .indices import bindex

_ADD_CACHE = {}
_MULTI_INDEX_CACHE = {}


def add(idxi, idxj, dim):
    """
    Bertran addition.

    Args:
        idxi (Tuple):
            Index in integer notation
        idxj (Tuple):
            Index in integer notation
        dim (int):
            The number of dimensions of the expansion.

    Examples:
        >>> print(chaospy.bertran.add(3, 3, 1))
        6
        >>> print(chaospy.bertran.add(3, 3, 2))
        10
    """
    key = idxi, idxj, dim
    if key in _ADD_CACHE:
        return _ADD_CACHE[key]

    idxi = multi_index(idxi, dim)
    idxj = multi_index(idxj, dim)
    out = single_index(tuple(i+j for i,j in zip(idxi, idxj)))

    _ADD_CACHE[key] = out
    return out


def terms(order, dim):
    """
    Count the number of polynomials in an expansion.

    Args:
        order (int):
            The upper order for the expansion.
        dim (int):
            The number of dimensions of the expansion.

    Returns:
        The number of terms in an expansion of upper order ``order`` and number
        of dimensions ``dim``.
    """
    return int(comb(order+dim, dim, exact=True))


def multi_index(idx, dim):
    """
    Single to multi-index using graded reverse lexicographical notation.

    Args:
        idx (int):
            Index in integer notation
        dim (int):
            The number of dimensions in the multi-index notation

    Returns (Tuple):
        Multi-index of ``idx`` with ``len(out) == dim``.

    Examples:
        >>> for idx in range(5):
        ...     print(chaospy.bertran.multi_index(idx, 3))
        (0, 0, 0)
        (1, 0, 0)
        (0, 1, 0)
        (0, 0, 1)
        (2, 0, 0)

    See Also:
        :func:`single_index`
    """
    key = idx, dim
    if key in _MULTI_INDEX_CACHE:
        return _MULTI_INDEX_CACHE[key]

    if not dim:
        out = ()

    elif idx == 0:
        out = (0, )*dim

    else:
        idxn = idxm = 0
        while terms(idxn, dim) <= idx:
            idxn += 1
        idx -= terms(idxn-1, dim)

        if idx == 0:
            out = (idxn,) + (0,)*(dim-1)
        else:
            while terms(idxm, dim-1) <= idx:
                idxm += 1
            out = (int(idxn-idxm),) + multi_index(idx, dim-1)

    _MULTI_INDEX_CACHE[key] = out
    return out


def single_index(idxm):
    """
    Multi-index to single integer notation.

    Uses graded reverse lexicographical notation.

    Parameters
    ----------
    idxm : numpy.ndarray
        Index in multi-index notation

    Returns
    -------
    idx : int
        Integer index of `idxm`

    Examples
    --------
    >>> for idx in range(3):
    ...     print(chaospy.bertran.single_index(numpy.eye(3)[idx]))
    1
    2
    3
    """
    if -1 in idxm:
        return 0
    order = int(sum(idxm))
    dim = len(idxm)
    if order == 0:
        return 0
    return terms(order-1, dim) + single_index(idxm[1:])


def rank(idx, dim):
    """Calculate the index rank according to Bertran's notation."""
    idxm = multi_index(idx, dim)
    out = 0
    while idxm[-1:] == (0,):
        out += 1
        idxm = idxm[:-1]
    return out


def parent(idx, dim, axis=None):
    """
    Parent node according to Bertran's notation.

    Parameters
    ----------
    idx : int
        Index of the child node.
    dim : int
        Dimensionality of the problem.
    axis : int
        Assume axis direction.

    Returns
    -------
    out : int
        Index of parent node with `j<=i`, and `j==i` iff `i==0`.
    axis : int
        Dimension direction the parent was found.
    """
    idxm = multi_index(idx, dim)
    if axis is None:
        axis = dim - numpy.argmin(1*(numpy.array(idxm)[::-1] == 0))-1

    if not idx:
        return idx, axis

    if idxm[axis] == 0:
        idxi = parent(parent(idx, dim)[0], dim)[0]
        while child(idxi+1, dim, axis) < idx:
            idxi += 1
        return idxi, axis

    out = numpy.array(idxm) - 1*(numpy.eye(dim)[axis])
    return single_index(out), axis


def child(idx, dim, axis):
    """
    Child node according to Bertran's notation.

    Parameters
    ----------
    idx : int
        Index of the parent node.
    dim : int
        Dimensionality of the problem.
    axis : int
        Dimension direction to define a child.
        Must have `0<=axis<dim`

    Returns
    -------
    out : int
        Index of child node with `out > idx`.

    Examples
    --------
    >>> print(chaospy.bertran.child(4, 1, 0))
    5
    >>> print(chaospy.bertran.child(4, 2, 1))
    8
    """
    idxm = multi_index(idx, dim)
    out = numpy.array(idxm) + 1*(numpy.eye(len(idxm))[axis])
    return single_index(out)


def olindex(order, dim):
    """
    Create an lexiographical sorted basis for a given order.

    Examples
    --------
    >>> chaospy.bertran.olindex(3, 2)
    array([[0, 3],
           [1, 2],
           [2, 1],
           [3, 0]])
    """
    idxm = [0]*dim
    out = []

    def _olindex(idx):
        """Recursive backend for olindex."""
        if numpy.sum(idxm) == order:
            out.append(idxm[:])
            return

        if idx == dim:
            return

        idxm_sum = numpy.sum(idxm)
        idx_saved = idxm[idx]

        for idxi in range(order - numpy.sum(idxm) + 1):

            idxm[idx] = idxi

            if idxm_sum < order:
                _olindex(idx+1)

            else:
                break
        idxm[idx] = idx_saved

    _olindex(0)
    return numpy.array(out)


def olindices(order, dim):
    """
    Create an lexiographical sorted basis for a given order.

    Examples:
        >>> chaospy.bertran.olindices(2, 2)
        array([[0, 0],
               [0, 1],
               [1, 0],
               [0, 2],
               [1, 1],
               [2, 0]])
    """
    indices = [olindex(o, dim) for o in range(order+1)]
    indices = numpy.vstack(indices)
    return indices



"""Cosine."""
import numpy

from ..baseclass import Dist
from .. import evaluation, approximation


class Cos(Dist):
    """
    Cosine.

    Args:
        dist (Dist): Distribution to perform transformation on.

    Example:
        >>> distribution = chaospy.Cos(chaospy.Uniform(0, 1))
        >>> print(distribution)
        Cos(Uniform(lower=0, upper=1))
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.6967 0.8253 0.9211 0.9801]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [1.394  1.771  2.5679 5.0335]
        >>> print(numpy.around(distribution.sample(4), 4))
        [0.9406 0.6333 0.9988 0.8689]
        >>> print(numpy.around(distribution.mom(1), 4))
        0.8413
    """

    def __init__(self, dist):
        assert isinstance(dist, Dist)
        Dist.__init__(self, dist=dist)

    def _pdf(self, x, dist, cache):
        """Probability density function."""
        output = evaluation.evaluate_density(dist, numpy.arccos(x), cache=cache)
        output /= numpy.where(numpy.isin(x, [-1, 1]), numpy.inf, numpy.sqrt(1-x*x))
        return output

    def _cdf(self, x, dist, cache):
        return 1-evaluation.evaluate_forward(dist, numpy.arccos(x), cache=cache)

    def _ppf(self, q, dist, cache):
        return numpy.cos(evaluation.evaluate_inverse(dist, 1-q, cache=cache))

    def _bnd(self, x, dist, cache):
        out = numpy.cos(evaluation.evaluate_bound(
            dist, numpy.arccos(x), cache=cache))
        return out[::-1]

    def _mom(self, x, dist, cache):
        return approximation.approximate_moment(self, x)

    def __len__(self):
        return len(self.prm["dist"])

    def __str__(self):
        return self.__class__.__name__ + "(" + str(self.prm["dist"]) + ")"

    def _fwd_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.cos(dist)
        return self

    def _inv_cache(self, cache):
        dist = evaluation.get_forward_cache(self.prm["dist"], cache)
        if not isinstance(dist, Dist):
            return numpy.arccos(dist)
        return self



"""
Create samples from the `Hammersley set`_.

The Hammersley set is equivalent to the Halton sequence, except for one
dimension is replaced with a regular grid.

Example usage
-------------

Standard usage::

    >>> distribution = chaospy.J(chaospy.Uniform(0, 1), chaospy.Uniform(0, 1))
    >>> samples = distribution.sample(3, rule="M")
    >>> print(numpy.around(samples, 4))
    [[0.75  0.125 0.625]
     [0.25  0.5   0.75 ]]
    >>> samples = distribution.sample(4, rule="M")
    >>> print(numpy.around(samples, 4))
    [[0.75  0.125 0.625 0.375]
     [0.2   0.4   0.6   0.8  ]]

.. _Hammersley set: https://en.wikipedia.org/wiki/Low-discrepancy_sequence#Hammersley_set
"""
import numpy

from .halton import create_halton_samples


def create_hammersley_samples(order, dim=1, burnin=-1, primes=()):
    """
    Create samples from the Hammersley set.

    For ``dim == 1`` the sequence falls back to Van Der Corput sequence.

    Args:
        order (int):
            The order of the Hammersley sequence. Defines the number of samples.
        dim (int):
            The number of dimensions in the Hammersley sequence.
        burnin (int):
            Skip the first ``burnin`` samples. If negative, the maximum of
            ``primes`` is used.
        primes (tuple):
            The (non-)prime base to calculate values along each axis. If
            empty, growing prime values starting from 2 will be used.

    Returns:
        (numpy.ndarray):
            Hammersley set with ``shape == (dim, order)``.
    """
    if dim == 1:
        return create_halton_samples(
            order=order, dim=1, burnin=burnin, primes=primes)
    out = numpy.empty((dim, order), dtype=float)
    out[:dim-1] = create_halton_samples(
        order=order, dim=dim-1, burnin=burnin, primes=primes)
    out[dim-1] = numpy.linspace(0, 1, order+2)[1:-1]
    return out



"""Ali-Mikhail-Haq copula."""
import numpy
from scipy import special

from .baseclass import Archimedean, Copula
from ..baseclass import Dist


class ali_mikhail_haq(Archimedean):
    """Ali-Mikhail-Haq copula."""

    def __init__(self, length, theta=.5, eps=1e-6):
        assert -1 <= theta < 1
        self.length = length
        Dist.__init__(self, th=float(theta), eps=eps)

    def __len__(self):
        return self.length

    def gen(self, x, th):
        return numpy.log((1-th*(1-x))/x)

    def igen(self, x, th):
        return (1-th)/(numpy.e**x-th)


class AliMikhailHaq(Copula):
    """
    Ali-Mikhail-Haq copula.

    Examples:
        >>> distribution = chaospy.Iid(chaospy.Uniform(), 2)
        >>> distribution = chaospy.AliMikhailHaq(distribution)
        >>> print(distribution)
        AliMikhailHaq(Iid(Uniform(lower=0, upper=1), 2), theta=0.5)
        >>> mesh = numpy.meshgrid(*[numpy.linspace(0, 1, 5)[1:-1]]*2)
        >>> print(numpy.around(distribution.inv(mesh), 4))
        [[[0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]
          [0.25   0.5    0.75  ]]
        <BLANKLINE>
         [[0.2044 0.2634 0.3175]
          [0.4326 0.5099 0.5703]
          [0.6939 0.7533 0.7937]]]
        >>> print(numpy.around(distribution.fwd(distribution.inv(mesh)), 4))
        [[[0.25 0.5  0.75]
          [0.25 0.5  0.75]
          [0.25 0.5  0.75]]
        <BLANKLINE>
         [[0.25 0.25 0.25]
          [0.5  0.5  0.5 ]
          [0.75 0.75 0.75]]]
        >>> print(numpy.around(distribution.pdf(distribution.inv(mesh)), 4))
        [[1.1636 0.9937 0.9088]
         [1.0285 1.0267 1.0631]
         [0.8882 1.0238 1.1705]]
        >>> print(numpy.around(distribution.sample(4), 4))
        [[0.6536 0.115  0.9503 0.4822]
         [0.8886 0.1432 0.0725 0.4046]]
        >>> print(numpy.around(distribution.mom((1, 2)), 4))
        0.1822
    """

    def __init__(self, dist, theta=0.5, eps=1e-6):
        """
        Args:
            dist (Dist) : The Distribution to wrap
            theta (float) : Copula parameter
        """
        self._repr = {"theta": theta}
        trans = ali_mikhail_haq(len(dist), theta, eps)
        return Copula.__init__(self, dist=dist, trans=trans)



"""Laplace Probability Distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add



class laplace(Dist):
    """Laplace Probability Distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return numpy.e**-numpy.abs(x)/2

    def _cdf(self, x):
        return (1+numpy.sign(x)*(1-numpy.e**-abs(x)))/2

    def _mom(self, k):
        return .5*special.factorial(k)*(1+(-1)**k)

    def _ppf(self, x):
        return numpy.where(x>.5, -numpy.log(2*(1-x)), numpy.log(2*x))

    def _bnd(self, x):
        return -32., 32.

    def _ttr(self, k):
        from ...quadrature import quad_fejer, discretized_stieltjes
        q1, w1 = quad_fejer(500, (-32, 0))
        q2, w2 = quad_fejer(500, (0, 32))
        q = numpy.concatenate([q1,q2], 1)
        w = numpy.concatenate([w1,w2])*self.pdf(q[0])

        coeffs, _, _ = discretized_stieltjes(k, q, w)
        return coeffs[:, 0, -1]


class Laplace(Add):
    R"""
    Laplace Probability Distribution

    Args:
        mu (float, Dist) : Mean of the distribution.
        scale (float, Dist) : Scaleing parameter. scale > 0.

    Examples:
        >>> distribution = chaospy.Laplace(2, 2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.1674 1.5537 2.4463 3.8326]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.1 0.2 0.2 0.1]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 2.734  -0.9392  6.6165  1.9275]
        >>> print(numpy.around(distribution.mom(1), 4))
        2.0
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[ 2.      2.      2.    ]
         [ 8.     39.9995 86.4011]]
    """
    def __init__(self, mu=0, scale=1):
        self._repr = {"mu": mu, "scale": scale}
        Add.__init__(self, left=laplace()*scale, right=mu)



"""
Gauss-Kronrod quadrature is an adaptive method for Gaussian quadrature rule. It
builds on top of other quadrature rules by extending "pure" Gaussian quadrature
rules with extra abscissas and new weights such that already used abscissas can
be reused. For more details, see `Wikipedia article`_.

For each order ``N`` taken with ordinary Gaussian quadrature, Gauss-Kronrod
will create ``2N+1`` abscissas where all of the ``N`` "old" abscissas are all
interlaced between the "new" ones.

The algorithm is well suited for any Jacobi scheme, i.e. quadrature involving
Uniform or Beta distribution, and might work on others as well. However, it
will not work everywhere. For example `Kahaner and Monegato`_ showed that
higher order Gauss-Kronrod quadrature for Gauss-Hermite and Gauss-Laguerre does
not exist.

.. _Wikipedia article: https://en.wikipedia.org/wiki/Gauss%E2%80%93Kronrod_quadrature_formula
.. _Kahaner and Monegato: https://link.springer.com/article/10.1007/BF01590820

Example usage
-------------

Generate Gauss-Kronrod quadrature rules::

    >>> distribution = chaospy.Beta(2, 2, lower=-1, upper=1)
    >>> for order in range(4):  # doctest: +NORMALIZE_WHITESPACE
    ...     X, W = chaospy.generate_quadrature(
    ...         order, distribution, rule="gauss_kronrod")
    ...     print(numpy.around(X, 2), numpy.around(W, 2))
    [[-0.65 -0.    0.65]] [0.23 0.53 0.23]
    [[-0.82 -0.45  0.    0.45  0.82]] [0.07 0.26 0.34 0.26 0.07]
    [[-0.89 -0.65 -0.34 -0.    0.34  0.65  0.89]]
     [0.03 0.12 0.22 0.26 0.22 0.12 0.03]
    [[-0.93 -0.77 -0.54 -0.29 -0.    0.29  0.54  0.77  0.93]]
     [0.01 0.06 0.13 0.19 0.22 0.19 0.13 0.06 0.01]

Compare Gauss-Kronrod builds on top of Gauss-Legendre quadrature to pure
Gauss-Legendre::

    >>> distribution = chaospy.Uniform(-1, 1)
    >>> for order in range(5):
    ...     Xl, Wl = chaospy.generate_quadrature(
    ...         order, distribution, rule="gaussian")
    ...     Xk, Wk = chaospy.generate_quadrature(
    ...         order, distribution, rule="gauss_kronrod")
    ...     print(numpy.around(Xl, 2), numpy.around(Xk[:, 1::2], 2))
    [[0.]] [[-0.]]
    [[-0.58  0.58]] [[-0.58  0.58]]
    [[-0.77 -0.    0.77]] [[-0.77 -0.    0.77]]
    [[-0.86 -0.34  0.34  0.86]] [[-0.86 -0.34  0.34  0.86]]
    [[-0.91 -0.54  0.    0.54  0.91]] [[-0.91 -0.54 -0.    0.54  0.91]]

Gauss-Kronrod build on top of Gauss-Hermite quadrature::

    >>> distribution = chaospy.Normal(0, 1)
    >>> for order in range(2):
    ...     Xl, Wl = chaospy.generate_quadrature(
    ...         order, distribution, rule="gaussian")
    ...     Xk, Wk = chaospy.generate_quadrature(
    ...         order, distribution, rule="gauss_kronrod")
    ...     print(numpy.around(Xl, 2), numpy.around(Xk, 2))
    [[0.]] [[-1.73  0.    1.73]]
    [[-1.  1.]] [[-2.45 -1.    0.    1.    2.45]]

Applying Gauss-Kronrod to Gauss-Hermite quadrature, for an order known to not
exist::

    >>> chaospy.generate_quadrature(  # doctest: +IGNORE_EXCEPTION_DETAIL
    ...     5, distribution, rule="gauss_kronrod")
    Traceback (most recent call last):
        ...
    numpy.linalg.LinAlgError: \
Invalid recurrence coefficients can not be used for constructing Gaussian quadrature rule

Multivariate support::

    >>> distribution = chaospy.J(
    ...     chaospy.Uniform(0, 1), chaospy.Beta(4, 5))
    >>> X, W = chaospy.generate_quadrature(
    ...     1, distribution, rule="gauss_kronrod")
    >>> print(numpy.around(X, 3))  # doctest: +NORMALIZE_WHITESPACE
    [[0.037 0.037 0.037 0.037 0.037 0.211 0.211 0.211 0.211
      0.211 0.5   0.5   0.5   0.5   0.5   0.789 0.789 0.789
      0.789 0.789 0.963 0.963 0.963 0.963 0.963]
     [0.144 0.297 0.444 0.612 0.796 0.144 0.297 0.444 0.612
      0.796 0.144 0.297 0.444 0.612 0.796 0.144 0.297 0.444
      0.612 0.796 0.144 0.297 0.444 0.612 0.796]]
    >>> print(numpy.around(W, 3))  # doctest: +NORMALIZE_WHITESPACE
    [0.006 0.027 0.035 0.026 0.004 0.016 0.067 0.086 0.065
     0.011 0.02  0.085 0.11  0.083 0.014 0.016 0.067 0.086
     0.065 0.011 0.006 0.027 0.035 0.026 0.004]

Sources
-------

Code is adapted from `quadpy`_, which adapted his code from `W. Gautschi`_.
Algorithm for calculating Kronrod-Jacobi matrices was first published in paper
by `D. P. Laurie`_.

.. _quadpy: https://github.com/nschloe/quadpy
.. _W. Gautschi: https://www.cs.purdue.edu/archives/2002/wxg/codes/OPQ.html
.. _D. P. Laurie: https://doi.org/10.1090/S0025-5718-97-00861-2
"""
from __future__ import division, print_function
import math

import numpy

from .recurrence import (
    construct_recurrence_coefficients, coefficients_to_quadrature)
from .combine import combine_quadrature


def quad_gauss_kronrod(
        order,
        dist,
        rule="fejer",
        accuracy=100,
        recurrence_algorithm="",
):
    """
    Generate the abscissas and weights in Gauss-Kronrod quadrature.

    Args:
        order (int):
            The order of the quadrature.
        dist (chaospy.distributions.baseclass.Dist):
            The distribution which density will be used as weight function.
        rule (str):
            In the case of ``lanczos`` or ``stieltjes``, defines the
            proxy-integration scheme.
        accuracy (int):
            In the case ``rule`` is used, defines the quadrature order of the
            scheme used. In practice, must be at least as large as ``order``.
        recurrence_algorithm (str):
            Name of the algorithm used to generate abscissas and weights. If
            omitted, ``analytical`` will be tried first, and ``stieltjes`` used
            if that fails.

    Returns:
        (numpy.ndarray, numpy.ndarray):
            abscissas:
                The quadrature points for where to evaluate the model function
                with ``abscissas.shape == (len(dist), N)`` where ``N`` is the
                number of samples.
            weights:
                The quadrature weights with ``weights.shape == (N,)``.

    Raises:
        ValueError:
            Error raised if Kronrod algorithm results in negative recurrence
            coefficients.

    Example:
        >>> distribution = chaospy.Uniform(-1, 1)
        >>> abscissas, weights = quad_gauss_kronrod(3, distribution)
        >>> print(numpy.around(abscissas, 3))
        [[-0.977 -0.861 -0.64  -0.34   0.     0.34   0.64   0.861  0.977]]
        >>> print(numpy.around(weights, 3))
        [0.031 0.085 0.133 0.163 0.173 0.163 0.133 0.085 0.031]
    """
    assert not rule.startswith("gauss"), "recursive Gaussian quadrature call"

    length = int(numpy.ceil(3*(order+1) / 2.0))
    coefficients = construct_recurrence_coefficients(
        length, dist, rule, accuracy, recurrence_algorithm)

    coefficients = [kronrod_jacobi(order+1, coeffs) for coeffs in coefficients]

    abscissas, weights = coefficients_to_quadrature(coefficients)
    return combine_quadrature(abscissas, weights)


def kronrod_jacobi(order, coeffs):
    """
    Create the three-terms-recursion coefficients resulting from the
    Kronrod-Jacobi matrix.

    Augment three terms recurrence coefficients to add extra Gauss-Kronrod
    terms.

    Args:
        order (int):
            Order of the Gaussian quadrature rule.
        coeffs (numpy.ndarray):
            Three terms recurrence coefficients of the Gaussian quadrature
            rule.

    Returns:
        Three terms recurrence coefficients of the Gauss-Kronrod quadrature
        rule.
    """
    assert len(coeffs[0]) == int(math.ceil(3*order/2.0))+1

    bound = int(math.floor(3*order/2.0))+1
    coeffs_a = numpy.zeros(2*order+1)
    coeffs_a[:bound] = coeffs[0][:bound]

    bound = int(math.ceil(3*order/2.0))+1
    coeffs_b = numpy.zeros(2*order+1)
    coeffs_b[:bound] = coeffs[1][:bound]

    sigma = numpy.zeros((2, order//2+2))
    sigma[1, 1] = coeffs_b[order+1]

    for idx in range(order-1):
        idy = numpy.arange((idx+1)//2, -1, -1)
        sigma[0, idy+1] = numpy.cumsum(
            (coeffs_a[idy+order+1]-coeffs_a[idx-idy])*sigma[1, idy+1]+
            coeffs_b[idy+order+1]*sigma[0, idy]-
            coeffs_b[idx-idy]*sigma[0, idy+1]
        )
        sigma = numpy.roll(sigma, 1, axis=0)

    sigma[0, 1:order//2+2] = sigma[0, :order//2+1]
    for idx in range(order-1, 2*order-2):
        idy = numpy.arange(idx-order+1, (idx-1)//2+1)
        j = order-1-idx+idy
        sigma[0, j+1] = numpy.cumsum(
            -(coeffs_a[idy+order+1]-coeffs_a[idx-idy])*sigma[1, j+1]-
            coeffs_b[idy+order+1]*sigma[0, j+1]+
            coeffs_b[idx-idy]*sigma[0, j+2]
        )
        j = j[-1]
        idy = (idx+1)//2
        if idx%2 == 0:
            coeffs_a[idy+order+1] = (
                coeffs_a[idy]+
                (sigma[0, j+1]-coeffs_b[idy+order+1]*sigma[0, j+2])/
                sigma[1, j+2]
            )
        else:
            coeffs_b[idy+order+1] = sigma[0, j+1]/sigma[0, j+2]
        sigma = numpy.roll(sigma, 1, axis=0)

    coeffs_a[2*order] = (coeffs_a[order-1]-
                         coeffs_b[2*order]*sigma[0, 1]/sigma[1, 1])
    return numpy.asfarray([coeffs_a, coeffs_b])



"""
Algorithm 3.4 of 'Numerical Optimization' by Jorge Nocedal and Stephen J.
Wright

This is based on the MATLAB code from Michael L. Overton <overton@cs.nyu.edu>:
http://cs.nyu.edu/overton/g22_opt/codes/cholmod.m
"""
import numpy
import scipy.sparse


def gill_king(mat, eps=1e-16):
    """
    Gill-King algorithm for modified cholesky decomposition.

    Args:
        mat (numpy.ndarray):
            Must be a non-singular and symmetric matrix.  If sparse, the result
            will also be sparse.
        eps (float):
            Error tolerance used in algorithm.


    Returns:
        (numpy.ndarray):
            Lower triangular Cholesky factor.

    Examples:
        >>> mat = [[4, 2, 1], [2, 6, 3], [1, 3, -.004]]
        >>> lowtri = gill_king(mat)
        >>> print(numpy.around(lowtri, 4))
        [[2.     0.     0.    ]
         [1.     2.2361 0.    ]
         [0.5    1.118  1.2264]]
        >>> print(numpy.around(numpy.dot(lowtri, lowtri.T), 4))
        [[4.    2.    1.   ]
         [2.    6.    3.   ]
         [1.    3.    3.004]]
    """
    if not scipy.sparse.issparse(mat):
        mat = numpy.asfarray(mat)
    assert numpy.allclose(mat, mat.T)

    size = mat.shape[0]
    mat_diag = mat.diagonal()
    gamma = abs(mat_diag).max()
    off_diag = abs(mat - numpy.diag(mat_diag)).max()

    delta = eps*max(gamma + off_diag, 1)
    beta = numpy.sqrt(max(gamma, off_diag/size, eps))

    lowtri = _gill_king(mat, beta, delta)

    return lowtri


def _gill_king(mat, beta, delta):
    """Backend function for the Gill-King algorithm."""
    size = mat.shape[0]

    # initialize d_vec and lowtri
    if scipy.sparse.issparse(mat):
        lowtri = scipy.sparse.eye(*mat.shape)
    else:
        lowtri = numpy.eye(size)

    d_vec = numpy.zeros(size, dtype=float)

    # there are no inner for loops, everything implemented with
    # vector operations for a reasonable level of efficiency
    for idx in range(size):
        if idx == 0:
            idz = []     # column index: all columns to left of diagonal
                        # d_vec(idz) doesn't work in case idz is empty
        else:
            idz = numpy.s_[:idx]

        djtemp = mat[idx, idx] - numpy.dot(
            lowtri[idx, idz], d_vec[idz]*lowtri[idx, idz].T)
        # C(idx, idx) in book

        if idx < size - 1:
            idy = numpy.s_[idx+1:size]
            # row index: all rows below diagonal
            ccol = mat[idy, idx] - numpy.dot(
                lowtri[idy, idz], d_vec[idz]*lowtri[idx, idz].T)
            # C(idy, idx) in book
            theta = abs(ccol).max()
            # guarantees d_vec(idx) not too small and lowtri(idy, idx) not too
            # big in sufficiently positive definite case, d_vec(idx) = djtemp
            d_vec[idx] = max(abs(djtemp), (theta/beta)**2, delta)
            lowtri[idy, idx] = ccol/d_vec[idx]

        else:
            d_vec[idx] = max(abs(djtemp), delta)

    # convert to usual output format: replace lowtri by lowtri*sqrt(D) and
    # transpose
    for idx in range(size):
        lowtri[:, idx] = lowtri[:, idx]*numpy.sqrt(d_vec[idx])
        # lowtri = lowtri*diag(sqrt(d_vec)) bad in sparse case

    return lowtri



"""Total Sobol sensitivity index."""
import numpy

from ..conditional import E_cond
from ..variance import Var


def Sens_t(poly, dist, **kws):
    """
    Variance-based decomposition
    AKA Sobol' indices

    Total effect sensitivity index

    Args:
        poly (Poly):
            Polynomial to find first order Sobol indices on.
        dist (Dist):
            The distributions of the input used in ``poly``.

    Returns:
        (numpy.ndarray) :
            First order sensitivity indices for each parameters in ``poly``,
            with shape ``(len(dist),) + poly.shape``.

    Examples:
        >>> x, y = chaospy.variable(2)
        >>> poly = chaospy.Poly([1, x, y, 10*x*y])
        >>> dist = chaospy.Iid(chaospy.Uniform(0, 1), 2)
        >>> indices = chaospy.Sens_t(poly, dist)
        >>> print(indices)
        [[0.         1.         0.         0.57142857]
         [0.         0.         1.         0.57142857]]
    """
    dim = len(dist)
    if poly.dim < dim:
        poly = chaospy.poly.setdim(poly, len(dist))

    zero = [1]*dim
    out = numpy.zeros((dim,) + poly.shape, dtype=float)
    V = Var(poly, dist, **kws)
    for i in range(dim):
        zero[i] = 0
        out[i] = ((V-Var(E_cond(poly, zero, dist, **kws), dist, **kws)) /
                  (V+(V == 0))**(V!=0))
        zero[i] = 1
    return out



"""Hyperbolic secant distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add
from .deprecate import deprecation_warning


class hyperbolic_secant(Dist):
    """Hyperbolic secant distribution."""

    def __init__(self):
        Dist.__init__(self)

    def _pdf(self, x):
        return .5*numpy.cosh(numpy.pi*x/2.)**-1

    def _cdf(self, x):
        return 2/numpy.pi*numpy.arctan(numpy.e**(numpy.pi*x/2.))

    def _ppf(self, q):
        return 2/numpy.pi*numpy.log(numpy.tan(numpy.pi*q/2.))

    def _mom(self, k):
        shape = k.shape
        output = numpy.abs([special.euler(k_)[-1] for k_ in k.flatten()])
        return output.reshape(shape)

    def _bnd(self, x):
        return self._ppf(1e-10), self._ppf(1-1e-10)


class HyperbolicSecant(Add):
    """
    Hyperbolic secant distribution

    Args:
        loc (float, Dist) : Location parameter
        scale (float, Dist) : Scale parameter

    Examples:
        >>> distribution = chaospy.HyperbolicSecant(2, 2)
        >>> print(distribution)
        HyperbolicSecant(loc=2, scale=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [0.5687 1.5933 2.4067 3.4313]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.1469 0.2378 0.2378 0.1469]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 2.6397 -0.1648  5.2439  1.9287]
        >>> print(numpy.around(distribution.mom(1), 4))
        2.0
        >>> print(numpy.around(distribution.ttr([1, 2, 3]), 4))
        [[ 2.      2.      2.    ]
         [ 4.     16.0001 35.9976]]
    """
    def __init__(self, loc=0, scale=1):
        self._repr = {"loc": loc, "scale": scale}
        Add.__init__(self, left=hyperbolic_secant()*scale, right=loc)


Hypgeosec = deprecation_warning(HyperbolicSecant, "Hypgeosec")



"""Pareto type 1 distribution."""
import numpy
from scipy import special

from ..baseclass import Dist
from ..operators.addition import Add


class pareto1(Dist):
    """Pareto type 1 distribution."""

    def __init__(self, b):
        Dist.__init__(self, b=b)

    def _pdf(self, x, b):
        return b * x**(-b-1)

    def _cdf(self, x, b):
        return 1 -  x**(-b)

    def _ppf(self, q, b):
        return pow(1-q, -1.0/b)

    def _bnd(self, x, b):
        return 1.0, self._ppf(1-1e-10, b)


class Pareto1(Add):
    """
    Pareto type 1 distribution.

    Lower threshold at scale+loc and survival: x^-shape

    Args:
        shape (float, Dist): Tail index parameter
        scale (float, Dist): Scaling parameter
        shift (float, Dist): Location parameter

    Examples:
        >>> distribution = chaospy.Pareto1(2, 2, 2)
        >>> print(distribution)
        Pareto1(loc=2, scale=2, shape=2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(distribution.inv(q), 4))
        [4.2361 4.582  5.1623 6.4721]
        >>> print(numpy.around(distribution.fwd(distribution.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(distribution.pdf(distribution.inv(q)), 4))
        [0.7155 0.4648 0.253  0.0894]
        >>> print(numpy.around(distribution.sample(4), 4))
        [ 5.3981  4.126  10.9697  4.7794]
    """

    def __init__(self, shape=1, scale=1, loc=0):
        self._repr = {"shape": shape, "scale": scale, "loc": loc}
        Add.__init__(self, left=pareto1(shape)*scale, right=loc)
